from PyQt5 import QtWidgets, QtGui


class ImageWidget(QtWidgets.QWidget):
    def __init__(self, filename, parent=None):
        super().__init__(parent=parent)
        label = QtWidgets.QLabel()
        pixmap = QtGui.QPixmap(filename)
        label.setPixmap(pixmap)

        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.addWidget(label,0,1) #add the widget in the second colum
        layout.setColumnStretch(0,1) #set stretch of first
        layout.setColumnStretch(2,1) #and third column
        self.setLayout(layout)



from collections import OrderedDict

green_sea =    (22,  160, 133, )
nephritis =    (39,  174, 96,  )
belize_hole =  (41,  128, 185, )
amethyst =     (155, 89,  182, )
wet_asphalt =  (52,  73,  94,  )
orange =       (243, 156, 18,  )
sun_flower =   (241, 196, 15,  )
pumpkin =      (211, 84,  0,   )
pomegranate =  (192, 57,  43,  )
clouds =       (236, 240, 241, )
concrete =     (149, 165, 166, )

colors = OrderedDict([
    ('r'      ,      pomegranate),
    ('g'      ,      nephritis),
    ('b'      ,      belize_hole),
    ('c'      ,      green_sea),
    ('m'      ,      amethyst),
    ('y'      ,      sun_flower),
    ('k'      ,      wet_asphalt),
    ('w'      ,      clouds),
    ('o'      ,      orange),
    ('gr'     ,      concrete),
    ('red'    ,      pomegranate),
    ('green'  ,      nephritis),
    ('blue'   ,      belize_hole),
    ('cyan'   ,      green_sea),
    ('magenta',      amethyst),
    ('yellow' ,      sun_flower),
    ('black'  ,      wet_asphalt),
    ('white'  ,      clouds),
    ('orange' ,      orange),
    ('gray'   ,      concrete),
])

cyclic_colors = [
    colors['r'],
    colors['g'],
    colors['b'],
    colors['m'],
    colors['c'],
    colors['w'],
    colors['y'],
    colors['o'],
    colors['gr'],
]



"""The base exception class and subclasses for exceptions raised by NSpyre.

The InstrumentManagerWindow class is the main GUI window for viewing the live
state and settings of the hardware devices connected to an NSpyre InstrumentServer.
It is defined by the QtWidgets.QMainWindow subclass from Qt and consists of a
QTreeWidget and subsequent QTreeWidgetItem(s) for displaying the attributes of
each device located on each connected Instrument Server. It is also responsible
for creating and connecting all the callback functions to the device drivers'
PySignal.Signal(s) for updating it's own GUI QtWidgets.QtWidget(s) in real-time.
i.e. QComboBox, QLineEdit, SpinBox, etc.


  Typical usage example:

  app = app.NSpyreApp([sys.argv])
  pyqtgraph._connectCleanup()
  with ..inserv.gateway.InservGateway() as isg:
      window = main_window.InstrumentManagerWindow(isg)
      sys.exit(app.exec())

Copyright (c) 2020, Alexandre Bourassa, Michael Solomon, Jacob Feder
All rights reserved.

This work is licensed under the terms of the 3-Clause BSD license.
For a copy, see <https://opensource.org/licenses/BSD-3-Clause>.
"""
import logging

logger = logging.getLogger(__name__)

class NSpyreError(Exception):
    """Base class for all NSpyre exceptions."""
    def __init__(self, message='', exception=None, error_labels=None):
        super().__init__(message)
        self._message = message
        self._error_labels = set(error_labels or [])
        if exception:
            logger.exception(exception)

    def has_error_label(self, label):
        """Return True if this error contains the given label."""
        return label in self._error_labels

    def _add_error_label(self, label):
        """Add the given label to this error."""
        self._error_labels.add(label)

    def _remove_error_label(self, label):
        """Remove the given label from this error."""
        self._error_labels.discard(label)


class EntryNotFoundError(NSpyreError):
    """Raised for failures Exception for when a configuration file doesn't contain the desired
    entry"""
    def __init__(self, config_path, msg=None):
        if msg is None:
            msg = 'Config file was expected to contain parameter: [{}] ' \
                    'but it wasn\'t found'.format(' -> '.join(config_path))
        super().__init__(msg)
        self.config_path = config_path


class ConfigurationError(NSpyreError):
    """Raised for failures with loading configuration files."""


class SpyreletLoadError(NSpyreError):
    """Raised for failures while loading a Spyrelet."""


class SpyreletUnloadError(NSpyreError):
    """Raised for failures with unloading a Spyrelet."""


class SpyreletRunningError(NSpyreError):
    """Raised for failures with executing the main method of a Spyrelet."""


class OSNotSupportedError(NSpyreError):
    """Raise for failures with"""


class InservGatewayError(NSpyreError):
    """Raised for failures related to the InservGateway."""


class InstrumentServerError(NSpyreError):
    """Raised for failures related to the InstrumentServer Service."""


class InstrumentManagerError(NSpyreError):
    """Raised for failures related to the InstrumentManagerWindow QMainWindow."""



"""
This module defines basic constants and functions that will be used
throughout nspyre

Copyright (c) 2020, Alexandre Bourassa, Michael Solomon, Jacob Feder
All rights reserved.

This work is licensed under the terms of the 3-Clause BSD license.
For a copy, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

from pathlib import Path

from pint import get_application_registry


# create a pint registry universal to nspyre
# uses get_application_registry to check if one
# already exists before creating a new instance
ureg = get_application_registry()
Q_ = ureg.Quantity

# root directory of nspyre
NSPYRE_ROOT = Path(__file__).parent


def join_nspyre_path(path):
    """Return a full path from a path given relative to the nspyre root 
    directory"""
    return NSPYRE_ROOT / path


# config files
CLIENT_META_CONFIG_PATH = join_nspyre_path('config/client_meta_config.yaml')
SERVER_META_CONFIG_PATH = join_nspyre_path('config/server_meta_config.yaml')

# images
LOGO_PATH = str(join_nspyre_path('gui/images/spyre.png'))


# format for accessing instrument server devices from a client, where the first
# item is the server name and the second is the device name (both defined in the
# server config file)
INSERV_DEV_ACCESSOR = '{}/{}'
# mongodb replicaset name
MONGO_RS = 'NSpyreSet'
# all spyrelet databases in mongodb will be of this form
MONGO_SPYRELETS_KEY = 'spyrelet[{}]'
# mongodb instrument server databases will contain a special document
# that contains the instrument server settings
MONGO_SERVERS_SETTINGS_KEY = '_settings'
# All instrument server databases in mongodb will be of this form
MONGO_SERVERS_KEY = 'inserv[{}]'
# in ms
MONGO_CONNECT_TIMEOUT = 5000
# rpyc connection timeout in s (None for no timeout)
RPYC_CONN_TIMEOUT = 30
# rpyc send/receive timeout in s (don't set to None)
RPYC_SYNC_TIMEOUT = 30

# config file key for mongodb address
CONFIG_MONGO_ADDR_KEY = 'mongodb_addr'



import logging
import time

from itertools import cycle

import numpy as np

from nspyre.gui import LinePlotWidget, Plot1D, Plot2D, PlotFormatInit, PlotFormatUpdate, colors
from nspyre.spyrelet import Spyrelet

COLORS = cycle(colors.keys())


class SubSpyrelet(Spyrelet):
    REQUIRED_DEVICES = [
        'sg'
    ]

    PARAMS = {
        'iterations': {'type': int, 'positive': True},
        'amplitude': {'type': float, 'positive': True},
    }

    def main(self, iterations, amplitude):
        for i in self.progress(range(iterations)):
            self.sg.amplitude = amplitude
            time.sleep(0.1)
            self.acquire({
                'ind': i,
                'rand': np.random.uniform(1,2, 100),
                'freq': self.sg.frequency,
            })
    @Plot1D
    def last_rand(df, cache):
        last_rand = np.array(df.rand.iloc[-1])
        return {'rand': [np.arange(len(last_rand)), last_rand]}

    @Plot1D
    def avg_rand(df, cache):
        rand = np.array(list(df.rand.values))
        return {'rand': [df.ind.values, rand.mean(axis=1)]}

    @Plot2D
    def data_im(df, cache):
        im = np.array(list(df.rand.values))
        return im

    @PlotFormatUpdate(LinePlotWidget, ['last_rand'])
    def formatter(plot, df, cache):
        for item in plot.plot_item.listDataItems():
            item.setPen(colors[next(COLORS)])


class MyExperiment(Spyrelet):
    REQUIRED_DEVICES = [
        'sg',
        'osc'
    ]

    REQUIRED_SPYRELETS = {
        's2': SubSpyrelet
    }

    PARAMS = {
        'amplitude': {
                'type': float,
                'units':'V'},
        'fs': {
            'type': range,
            'units': 'GHz'},
    }

    def main(self, fs, amplitude):
        for i, f in enumerate(self.progress(fs)):
            self.sg.frequency = f
            self.call(self.s2, 100, amplitude)
            val = self.s2.data.rand.mean().mean()
            self.acquire({
                'ind': i,
                'f': f,
                'A': amplitude,
                'result': val,
            })

    def initialize(self, fs, amplitude):
        logging.info('initializing [{}]...'.format(self.name))

    def finalize(self, fs, amplitude):
        logging.info('finalizing [{}]...'.format(self.name))

    @Plot1D
    def plot_results(df, cache):
        return {'result': [df.ind.values, df.result.values]}

    @Plot1D
    def plot_f(df, cache):
        return {'f': [df.ind.values, df.f.values], 'f2':[df.ind.values, 2*df.f.values]}

    @PlotFormatInit(LinePlotWidget, ['plot_f', 'plot_results'])
    def init_formatter(plot):
        plot.xlabel = 'My x axis (a.u.)'
        plot.ylabel = 'Signal (in bananas)'

# if __name__=='__main__':
#     from nspyre.instrument_manager import Instrument_Manager
#     from nspyre.instrument_server import Instrument_Server_Client
#     mongodb_addrs = ["mongodb://localhost:27017/","mongodb://localhost:27018/"]
#     c = Instrument_Server_Client(ip='localhost', port='5556')
#     m = Instrument_Manager([c])
#     sg  = m.get('sg')
#     s2 = SubSpyrelet('s2', mongodb_addrs, m)
#     s1 = MyExperiment('myExp', mongodb_addrs, m, {'s2':s2})


from lantz.core import Driver, Feat, DictFeat

DIN_KEYS = range(1, 10)
DOUT_KEYS = range(1, 10)

AIN_KEYS = range(1, 10)
AOUT_KEYS = range(1, 10)

class FakeDAQ(Driver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._din = {k : False for k in DIN_KEYS}
        self._dout = {k : False for k in DOUT_KEYS}
        self._ain = {k : 0.0 for k in AIN_KEYS}
        self._aout = {k : 0.0 for k in AOUT_KEYS}

    # digital input
    @DictFeat(values={True, False}, keys=list(DIN_KEYS))
    def din(self, key):
        return self._din[key]
    def reset_din(self, val):
        self._din = {k : val for k in DIN_KEYS}
    def toggle_din(self, key):
        self._din[key] = not self._din[key]

    # digital output
    @DictFeat(values={True, False}, keys=list(DOUT_KEYS))
    def dout(self, key):
        return self._dout[key]
    @dout.setter
    def dout(self, key, value):
        self._dout[key] = value

    # analog input
    @DictFeat(units='V', keys=list(AIN_KEYS))
    def ain(self, key):
        return self._ain[key]

    # analog output
    @DictFeat(units='V', keys=list(AIN_KEYS))
    def aout(self, key):
        return self._aout[key]
    @dout.setter
    def aout(self, key, value):
        self._aout[key] = value



# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import codecs
import os
import pathlib
import re
import sys
sys.path.insert(0, os.path.abspath('.'))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    here = pathlib.Path(__file__).parent.resolve()
    with codecs.open(os.path.join(here, *parts), "rb", "utf-8") as f:
        return f.read()


def find_version(*file_paths):
    """
    Build a path from *file_paths* and search for a ``__version__``
    string inside.
    """
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# -- Project information -----------------------------------------------------

project = 'nspyre'
copyright = '2020, Alexandre Bourassa'
author = 'Alexandre Bourassa'

# The full version, including alpha/beta/rc tags
meta_path = os.path.join('..', 'src', 'nspyre', '__init__.py')
release = find_version(meta_path)


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.

needs_sphinx = '3.1.2'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax', # for math formulas
    'sphinx.ext.napoleon', # for numpy and google style docstrings
    #'sphinx_autodoc_typehints',
    #'sphinx.ext.viewcode',
    #'sphinx.ext.extlinks',
    #'sphinx_tabs.tabs',
    #'sphinx_automodapi.automodapi',
    #'jupyter_sphinx',
    'nbsphinx', # for .ipynb file support (i.e. jupyter notebooks)
    #'sphinxcontrib.bibtex', # for bibliographic references
    'sphinx.ext.imgconverter',
    'sphinxcontrib.rsvgconverter', # for SVG->PDF conversion in LaTeX output
    'sphinx_copybutton' # for adding 'copy to clipboard' buttons to all text/code boxes
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_theme_options = {
    #'canonical_url': '',
    #'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    #'vcs_pageview_mode': '',
    'style_nav_header_background': '#2b2b2b',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': False,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['custom.css']

#html_logo = 'images/logo.png'
#html_favicon = 'images/favicon.ico'

html_last_updated_fmt = '%b %d, %Y'
#'%Y/%m/%d'



from collections.abc import Iterable
import functools
import importlib
import inspect
import sys
import warnings

import numpy as np
import pymongo
import rpyc

from nspyre.definitions import Q_, MONGO_RS, CONFIG_MONGO_ADDR_KEY
from nspyre.config import get_config_param, load_config, load_meta_config

string_types = (type(b''), type(u''))


def deprecated(reason):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter('always', DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2
                )
                warnings.simplefilter('default', DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2
            )
            warnings.simplefilter('default', DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))


def register_quantity_brining(quantity_class):
    """pint has an associated unit registry, and Quantity objects 
    cannot be shared between registries. Because Quantity objects 
    passed from the client to server or vice versa have a different 
    unit registry, they must be converted to Quantity objects of the 
    local registry. RPyC serializes objects using "brine". We will 
    make a custom brine serializer for Quantity objects to properly 
    pack and unpack them using the local unit registry.
    For more details, see pint documentation and
    https://github.com/tomerfiliba-org/rpyc/blob/master/rpyc/core/brine.py

    quantity_class: the Quantity class object from the local pint registry
    """
    rpyc.core.brine.TAG_PINT_Q = b"\xFA"

    # function for serializing quantity objects
    @rpyc.core.brine.register(rpyc.core.brine._dump_registry,
                                type(quantity_class(1, 'V')))
    def _dump_quantity(obj, stream):
        stream.append(rpyc.core.brine.TAG_PINT_Q)
        quantity_tuple = obj.to_tuple()
        rpyc.core.brine._dump((float(quantity_tuple[0]), \
                                quantity_tuple[1]), stream)

    # function for deserializing quantity objects
    @rpyc.core.brine.register(rpyc.core.brine._load_registry,
                            rpyc.core.brine.TAG_PINT_Q)
    def _load_quantity(stream):
        q = quantity_class.from_tuple(rpyc.core.brine._load(stream))
        return q
    rpyc.core.brine.simple_types = rpyc.core.brine.simple_types.union(\
                                    frozenset([type(quantity_class(1, 'V'))]))


def load_class_from_str(class_str):
    """Load a python class object (available in the local scope)
    from a string"""
    class_name = class_str.split('.')[-1]
    module_name = class_str.replace('.' + class_name, '')
    if module_name in sys.modules:
        # we just need to reload it
        mod = importlib.reload(sys.modules[module_name])
    else:
        # load the module
        mod = importlib.import_module(module_name)
    return getattr(mod, class_name)


def load_class_from_file(file_path, class_name):
    """Load a python class object from an external python file"""
    file_dir = file_path.parent
    file_name = str(file_path.name)
    # remove .py extension
    file_name = file_name.split('.py')[0]
    # load the spyrelet class from its python file
    sys.path.append(str(file_dir))
    if file_name in sys.modules:
        loaded_module = importlib.reload(sys.modules[file_name])
    else:
        loaded_module = importlib.import_module(file_name)
    loaded_class = getattr(loaded_module, class_name)
    return loaded_class


def qt_set_trace():
    """Set a tracepoint in the Python debugger that works with Qt"""
    from PyQt5.QtCore import pyqtRemoveInputHook
    from pdb import set_trace
    pyqtRemoveInputHook()
    set_trace()


def cleanup_register(client):
    if type(client) is str:
        client = pymongo.MongoClient(client)
    db_list = client['Spyre_Live_Data'].list_collection_names()
    reg_list = [x['_id'] for x in client['Spyre_Live_Data']['Register'].find({},{'_id':True})]
    for name in reg_list:
        if not name in db_list:
            client['Spyre_Live_Data']['Register'].delete_one({'_id': name})


def custom_encode(d):
    out = dict()
    for k,val in d.items():
        if type(val) == Q_:
            out[k] = {'__type__':'Quantity', 'm':val.m, 'units':str(val.units)}
        elif type(val) == RangeDict:
            out[k] = {'__type__':'RangeDict'}
            out[k].update(custom_encode(val))
        elif type(val) == np.ndarray:
            out[k] = {'__type__':'ndarray', 'val':val.tolist()}
        else:
            out[k] = val
    return out


def custom_decode(d):
    if isinstance(d, list):
        out = []
        for val in d:
            if isinstance(val, Iterable) and '__type__' in val:
                if val['__type__'] == 'Quantity':
                    out.append(Q_(val['m'], val['units']))
                elif val['__type__'] == 'RangeDict':
                    out.append(custom_decode(val))
                elif val['__type__'] == 'ndarray':
                    out.append(np.array(val['val']))
            else:
                out.append(val)
        return out
    else:
        out = dict()
        for k,val in d.items():
            if isinstance(val, Iterable) and '__type__' in val:
                if val['__type__'] == 'Quantity':
                    out[k] = Q_(val['m'], val['units'])
                elif val['__type__'] == 'RangeDict':
                    out[k] = custom_decode(val)
                elif val['__type__'] == 'ndarray':
                    out[k] = np.array(val['val'])
            else:
                out[k] = val
        return out

# pymongo does not like the deprecated decorator, so it's not being used for now.
# @deprecated('Use a normal numpy.linspace, numpy.arange, or numpy.logspace. This will be removed in a future version of nspyre.')
class RangeDict(dict):
    FUNCS = {'linspace': np.linspace, 'arange': np.arange, 'logspace': np.logspace}

    def __init__(self, **initial_dict):
        self._verify_valid(initial_dict)
        dict.__setitem__(self, 'func', initial_dict.pop('func'))
        for k in initial_dict:
            self[k] = initial_dict[k]

    def _verify_valid(self, d):
        # Make sure there is a valid func argument
        if not 'func' in d or d['func'] not in self.FUNCS:
            raise ValueError('RangeDict must have a element <func> in: {}'.format(self.FUNCS.keys()))
        
        # Make sure all required args are define
        signature = inspect.signature(self.FUNCS[d['func']])
        for param in signature.parameters.keys():
            if signature.parameters[param].default == inspect._empty and not param in d:
                raise ValueError('RangeDict does not define required <{}> argument'.format(param))

    def __setitem__(self, key, val):
        if key == 'func':
            raise KeyError('Cannot redefine the func, you should reinstantiate with new parameters')
        elif key in inspect.signature(self.FUNCS[self['func']]).parameters:
            return dict.__setitem__(self, key, val)

    @property
    def array(self):
        d = self.copy()
        func_name = d.pop('func')
        units = None
        for p in d:
            if type(d[p])==Q_:
                if units is None:
                    units = d[p].units
                d[p] = d[p].to(units).m
        if units is None or units == Q_(1, 'dimensionless').units:
            return self.FUNCS[func_name](**d)
        else:
            return self.FUNCS[func_name](**d)*units


def get_mongo_client(mongodb_addr=None):
    if mongodb_addr is None:
        cfg_path = load_meta_config()
        cfg = load_config(cfg_path)
        mongodb_addr, _ = get_config_param(cfg, [CONFIG_MONGO_ADDR_KEY])
    return pymongo.MongoClient(mongodb_addr, replicaset=MONGO_RS)



import inspect
import itertools as it

from collections import OrderedDict

import numpy as np
import pyqtgraph as pg

from PyQt5 import QtWidgets, QtCore
from pyqtgraph import functions as fn
from scipy.optimize import curve_fit

from ..colormap import viridis
from ..colors import cyclic_colors, colors
from .code_editor import Scintilla_Code_Editor, Monokai_Python_Lexer
from .spinbox import SpinBox
from .splitter_widget import Splitter, SplitterOrientation

pg.setConfigOptions(imageAxisOrder='row-major')


class BasePlotWidget(QtWidgets.QWidget):

    def __init__(self, w=None, plot_item=None, parent=None):
        super().__init__(parent=parent)
        if w is None:
            w = pg.PlotWidget()
        if plot_item is None:
            plot_item = w.getPlotItem()
        self.w = w
        self.plot_item = plot_item
        self.xonbottom = True
        self.yonleft = True
        self.traces = dict()

        self._title = ''
        self._xlabel = ''
        self._ylabel = ''
        self.invertY = False
        self.invertX = False

        self.init_plot()

        return

    def init_plot(self):
        self.build_toolbox()

        splitter_config = {
            'main_w': self.w,
            'side_w': self.toolbox,
            'orientation': SplitterOrientation.vertical_left_button,
        }
        splitter = Splitter(**splitter_config)

        def show_event(ev):
            h = splitter.size().height()
            splitter.setSizes([1, 0])
            splitter.resize(h, h)

        # Collaspe the tools to start
        splitter.showEvent = show_event

        layout = QtWidgets.QGridLayout()
        layout.addWidget(splitter)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)
        return

    def build_toolbox(self):
        # create default tools
        self.crosshairs = CrosshairAddon(self.plot_item)

        # create toolbox and add tools
        self.toolbox = QtWidgets.QToolBox()
        self.toolbox.addItem(self.crosshairs, "Crosshairs")
        return

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, _title):
        self._title = _title
        self.plot_item.setTitle(title=_title)
        return

    @property
    def xlabel(self):
        return self._xlabel

    @xlabel.setter
    def xlabel(self, _xlabel):
        self._xlabel = _xlabel
        pos = 'bottom' if self.xonbottom else 'top'
        self.plot_item.setLabel(pos, _xlabel)
        return

    @property
    def ylabel(self):
        return self._ylabel

    @ylabel.setter
    def ylabel(self, _ylabel):
        self._ylabel = _ylabel
        pos = 'left' if self.yonleft else 'right'
        self.plot_item.setLabel(pos, _ylabel)
        return

    def clear(self):
        ypos = 'left' if self.yonleft else 'right'
        xpos = 'bottom' if self.xonbottom else 'top'
        self.plot_item.setLabels(**{'title':'', xpos:'', ypos:''})

    @property
    def invertY(self):
        return self._invertY

    @invertY.setter
    def invertY(self, invert):
        self.plot_item.getViewBox().invertY(invert)
        self._invertY = invert
        return

    @property
    def invertX(self):
        return self._invertX

    @invertX.setter
    def invertX(self, invert):
        self.plot_item.getViewBox().invertX(invert)
        self._invertX = invert
        return

    def generate_meta(self, **kwargs):
        d = {'xlabel':self._xlabel, 'ylabel':self._ylabel, 'title':self._title, 'invertX':self._invertX, 'invertY':self._invertY}
        d.update(kwargs)
        return d


    def load_meta(self, meta, params=['xlabel', 'ylabel', 'title', 'invertX', 'invertY']):
        for key in params:
            if key in meta and hasattr(self, key):
                setattr(self, key, meta[key])


class HeatmapPlotWidget(BasePlotWidget):

    def __init__(self, parent=None, cmap=None):
        plot_item = pg.PlotItem(enableMouse=False)
        w = pg.ImageView(view=plot_item)
        super().__init__(parent=parent, w=w, plot_item=plot_item)
        w.ui.roiBtn.clicked.connect(self._set_roi_pos)
        self.grid(False)
        gradient = self.w.ui.histogram.gradient
        if cmap is None:
            cmap = viridis
        gradient.setColorMap(cmap)
        for tick in gradient.ticks:
            tick.hide()
        gradient.setFixedWidth(300)
        # disable gradient editing
        gradient.mouseClickEvent = lambda ev: None
        self.add_plotting_options()
        self._pos = None
        self._scale = None

        self.invertY = True
        return

    def _set_roi_pos(self):
        #Tries to place the ROI in a good spot when clicking the button
        r = np.array(self.plot_item.getViewBox().viewRange())
        pos = r.mean(axis=1)
        size = np.diff(r, axis=1)[:,0]
        if any(abs(pos-self.w.roi.pos())>abs(size)):
            self.w.roi.setPos(pos)
        if any(2*size<self.w.roi.size()):
            self.w.roi.setSize(size/2)


    def add_plotting_options(self):
        layout = QtWidgets.QVBoxLayout()
        self.plot_opts_checkboxes = dict()
        for k,val in [('autoHistogramRange',False), ('autoLevels',True), ('autoRange',True)]:
            self.plot_opts_checkboxes[k] = w = QtWidgets.QCheckBox(k)
            w.setChecked(val)
            layout.addWidget(w)
        
        #Add aspect lock option
        self.plot_opts_checkboxes['aspectLocked'] = w = QtWidgets.QCheckBox('aspectLocked')
        w.setChecked(True)
        layout.addWidget(w)
        w.stateChanged.connect(lambda x: setattr(self, 'aspectLocked', self.plot_opts_checkboxes['aspectLocked'].isChecked()))


        tool_w = QtWidgets.QWidget()
        tool_w.setLayout(layout)
        self.toolbox.addItem(tool_w, "Plotting Options")


    def grid(self, toggle, alpha=0.4):
        self.plot_item.showGrid(x=toggle, y=toggle, alpha=alpha)
        return

    @property
    def aspectLocked(self):
        return self.w.getView().state['aspectLocked']

    @aspectLocked.setter
    def aspectLocked(self, lock):
        return self.w.getView().setAspectLocked(lock)

    @property
    def im_pos(self):
        return self._pos

    @im_pos.setter
    def im_pos(self, pos):
        self._pos = pos

    @property
    def im_scale(self):
        return self._scale

    @im_scale.setter
    def im_scale(self, scale):
        self._scale = scale

    def set(self, im):
        self.w.setImage(im, pos=self._pos, scale=self._scale,
                        autoRange=self.plot_opts_checkboxes['autoRange'].isChecked(),
                        autoLevels=self.plot_opts_checkboxes['autoLevels'].isChecked(),
                        autoHistogramRange=self.plot_opts_checkboxes['autoHistogramRange'].isChecked(),)
        return

    def get(self):
        return self.w.getImageItem().image


class LinePlotWidget(BasePlotWidget):

    plots_updated = QtCore.pyqtSignal(list)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.legend = self.w.addLegend()
        self.grid(True)
        self._colors = it.cycle(cyclic_colors)
        self.install_fitter()
        return

    def install_fitter(self):
        self.fitter = FitterWidget(self.w)
        self.fitter.traces = self.traces
        self.toolbox.addItem(self.fitter, "Fitter")
        return

    def grid(self, toggle, alpha=0.4):
        self.w.showGrid(x=toggle, y=toggle, alpha=alpha)
        return

    def plot(self, tracename, **kwargs):
        symbol_pen = kwargs.get('symbol_pen', pg.mkPen(color=(255, 255, 255, 100)))
        symbol_brush = kwargs.get('symbol_brush', pg.mkBrush(color=(255, 255, 255, 100)))
        symbol_size = kwargs.get('symbol_size', 5)
        symbol = kwargs.get('symbol', 's')
        pen = kwargs.get('pen', pg.mkPen(color=(next(self._colors) + (200,)), width=1))
        antialias = kwargs.get('antialias', False)
        trace = self.w.plot(
            symbol=symbol,
            symbolSize=symbol_size,
            symbolPen=symbol_pen,
            symbolBrush=symbol_brush,
            pen=pen,
            antialias=antialias,
        )
        trace.curve.setClickable(True)
        trace_err = None
        self.legend.addItem(trace, tracename)
        self.traces[tracename] = trace, trace_err
        self.fitter.update_traces()
        return

    def remove_trace(self, tracename):
        for item in self.traces.pop(tracename):
            self.plot_item.removeItem(item)
        self.legend.removeItem(tracename)
        self.fitter.update_traces()

    def clear(self):
        super().clear()
        for tracename in list(self.traces.keys()):
            self.remove_trace(tracename)
        self._colors = it.cycle(cyclic_colors)

    def set(self, tracename, **kwargs):
        if tracename not in self.traces:
            self.plot(tracename, **kwargs)
        trace, trace_err = self.traces[tracename]
        data = kwargs.pop('data', None)
        xs = kwargs.pop('xs', None)
        ys = kwargs.pop('ys', None)
        yerrs = kwargs.pop('yerrs', None)
        if not any(item is None for item in [data, xs, ys]):
            raise ValueError('No plot points supplied (either data or xs and ys must be given)')
        if data is not None:
            xs, ys = list(zip(*data))
        if isinstance(xs, list):
            xs = np.array(xs)
        elif not isinstance(xs, np.ndarray):
            xs = xs.values
        if isinstance(ys, list):
            ys = np.array(ys)
        elif not isinstance(ys, np.ndarray):
            ys = ys.values
        trace.setData(x=xs, y=ys, **kwargs)
        if yerrs is not None:
            if trace_err is None:
                error_bar_params = {
                    'x': xs,
                    'y': ys,
                    'top': 0,
                    'bottom': 0,
                    'beam': 0.0,
                    'pen': pg.mkPen(color=(255, 255, 255, 80)),
                }
                trace_err = pg.ErrorBarItem(**error_bar_params)
                trace_err.setZValue(-999)
                self.w.addItem(trace_err)
                self.traces[tracename] = trace, trace_err
            ylines = list(zip(*[(yerr / 2, yerr / 2) if not np.isnan(yerr) else (0, 0) for y, yerr in zip(ys, yerrs)]))
            if ylines:
                ybottoms, ytops = ylines
                trace_err.setData(x=xs, y=ys, top=ytops, bottom=ybottoms, beam=0.0)
        return

    def get(self, tracename):
        pditem, _ = self.traces[tracename]
        x, y = pditem.getData()
        if x is None or y is None:
            return None, None
        else:
            return x.astype(np.float), y.astype(np.float)

    def __iter__(self):
        for tracename in self.traces:
            yield tracename

    def __getitem__(self, tracename):
        return self.get(tracename)

class FastImageWidget(BasePlotWidget):

    def __init__(self, parent=None):
        graphic_view = pg.GraphicsView()
        plot_item = pg.PlotItem(enableMouse=False)
        self.img_item = pg.ImageItem()
        plot_item.addItem(self.img_item)
        graphic_view.setCentralItem(plot_item)
        super().__init__(parent=parent, w=graphic_view, plot_item=plot_item)

    def set(self, image=None, autoLevels=None, **kwargs):
        self.img_item.setImage(image=image, autoLevels=autoLevels, **kwargs)


# ----------
# Crosshairs
# ----------
class Crosshair(QtCore.QObject):
    sigPositionChanged = QtCore.pyqtSignal(object)
    sigPositionChangeFinished = QtCore.pyqtSignal(object)

    def __init__(self, plot_item, pos, **kwargs):
        super().__init__()
        self.kwargs = kwargs
        self.moving = False
        self.hovering = False

        pen = fn.mkPen((255, 255, 0, 127))
        self.vLine = pg.InfiniteLine(angle=90, pen=pen, movable=False)
        self.hLine = pg.InfiniteLine(angle=0,  pen=pen, movable=False)
        self.vLine.hoverEvent, self.hLine.hoverEvent = self.hoverEvent, self.hoverEvent
        self.vLine.mouseDragEvent, self.hLine.mouseDragEvent = self.mouseDragEvent, self.mouseDragEvent

        self.center_dot = pg.ScatterPlotItem(pos=[pos], pen=fn.mkPen((255,0,0, 127)), brush=(255,0,0), symbol='o', size=3)

        plot_item.addItem(self.vLine, ignoreBounds=True)
        plot_item.addItem(self.hLine, ignoreBounds=True)
        plot_item.addItem(self.center_dot, ignoreBounds=True)
        self.plot_item = plot_item
        self.set_pos(pos)

    def set_pos(self, pos, emit_sig=True):
        if isinstance(pos, QtCore.QPointF):
            self.pos = [pos.x(), pos.y()]
        else:
            self.pos = list(pos)
        self.vLine.setPos(self.pos[0])
        self.hLine.setPos(self.pos[1])
        self.center_dot.setData(pos=[pos])
        if emit_sig: self.sigPositionChanged.emit(self.get_pos())

    def mouseDragEvent(self, ev):
        if ev.button() == QtCore.Qt.LeftButton:
            if ev.isStart():
                self.moving = True
            ev.accept()

            if not self.moving:
                return
            self.set_pos(self.plot_item.vb.mapSceneToView(ev.scenePos()))
            if ev.isFinish():
                self.moving = False
                self.sigPositionChangeFinished.emit(self.get_pos())

    def hoverEvent(self, ev):
        if (not ev.isExit()) and ev.acceptDrags(QtCore.Qt.LeftButton):
            self.hovering = True
            for line in [self.vLine, self.hLine]: line.currentPen = fn.mkPen(255, 0,0)
        else:
            self.hovering = False
            for line in [self.vLine, self.hLine]: line.currentPen = line.pen
        for line in [self.vLine, self.hLine]:
            line.update()

    def get_pos(self):
        return self.pos

    def delete(self):
        self.hLine.deleteLater()
        self.vLine.deleteLater()
        self.center_dot.deleteLater()


class CrosshairAddon(QtWidgets.QWidget):
    sigCrosshairAdded = QtCore.pyqtSignal(object)
    sigCrosshairRemoved = QtCore.pyqtSignal(object)
    def __init__(self, plot_item, **kwargs):
        super().__init__(**kwargs)
        self.plot_item = plot_item
        self.cross_list = list()
        self._spinbox_decimals = 4

        self.build_ui()


    @property
    def spinbox_decimals(self):
        return self._spinbox_decimals

    @spinbox_decimals.setter
    def spinbox_decimals(self, val):
        if self._spinbox_decimals != val:
            for r in range(self.table.rowCount()):
                self.table.cellWidget(r,0).setDecimals(val)
                self.table.cellWidget(r,1).setDecimals(val)
            self._spinbox_decimals = val

    def build_ui(self):
        self.table = QtWidgets.QTableWidget()
        self.table.setColumnCount(3)
        self.table.setHorizontalHeaderLabels(['x','y','Delete'])

        #Add control (for now just an add button)
        add_btn = QtWidgets.QPushButton('+ Add')
        def add():
            r = np.array(self.plot_item.getViewBox().viewRange())
            self.add_crosshair(r.mean(axis=1))
        add_btn.clicked.connect(lambda: add())

        #Add a decimal precision box
        decimal_input = SpinBox(value=self.spinbox_decimals, minStep=1, dec=False, int=True, bounds=(0, None), step=1)
        decimal_input.valueChanged.connect(lambda x: setattr(self, 'spinbox_decimals', decimal_input.value()))

        ctrl_layout = QtWidgets.QFormLayout()
        ctrl_layout.addRow('Add Crosshair', add_btn)
        ctrl_layout.addRow('Floating point precision', decimal_input)
        ctrl_widget = QtWidgets.QWidget()
        ctrl_widget.setLayout(ctrl_layout)


        layout = QtWidgets.QGridLayout()
        layout.addWidget(ctrl_widget)
        layout.addWidget(self.table)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def add_crosshair(self, pos, **kwargs):
        # Add the Crosshair to the list
        cross = Crosshair(self.plot_item, pos, **kwargs)
        self.cross_list.append(cross)

        # Add the table entry
        row = len(self.cross_list)-1
        self.table.insertRow(row)

        # Add the x,y widgets
        def update_pos(axis, value):
            if cross.moving:
                return
            cur = cross.get_pos()
            if axis==0:
                cross.set_pos([value, cur[1]], emit_sig=False)
            elif axis==1:
                cross.set_pos([cur[0], value], emit_sig=False)
        def lambda_gen(axis):
            return lambda obj: update_pos(axis, obj.value())

        for i in range(2):
            w = SpinBox(value = pos[i], dec=True, decimals=self.spinbox_decimals)
            self.table.setCellWidget(row, i, w)
            w.sigValueChanged.connect(lambda_gen(i))


        # Add a remove button
        btn = QtWidgets.QPushButton('X')
        btn.clicked.connect(lambda: self.remove_crosshair(cross))
        self.table.setCellWidget(row, 2, btn)

        # Link the position of the cross to the numbers in the table
        cross.sigPositionChanged.connect(lambda: self.update_table_entry(cross))
        self.sigCrosshairAdded.emit(cross)


    def _find_index(self, cross):
        for i in range(len(self.cross_list)):
            if self.cross_list[i] == cross:
                return i

    def update_table_entry(self, cross):
        row = self._find_index(cross)
        self.table.cellWidget(row, 0).setValue(cross.get_pos()[0])
        self.table.cellWidget(row, 1).setValue(cross.get_pos()[1])

    def remove_crosshair(self, cross):
        index = self._find_index(cross)
        self.table.removeRow(index)
        cross.delete()
        self.cross_list.pop(index)
        self.sigCrosshairRemoved.emit(index)


    def __getitem__(self, k):
        return self.cross_list[k].get_pos()

    def __iter__(self):
        for cross in self.cross_list:
            yield cross

    def __len__(self):
        return len(self.cross_list)


class FitterWidget(QtWidgets.QWidget):

    EXTRA_TEXT = """
# Fitter rules:
#     - Fit function name is "fit_func"
#     - Default values to arguments will be 
#       passed as start values for the fitter
#     - First argument of "fit_func" must be "xs"
#     - Have fun!

import numpy as np
"""

    DEFAULT_FUNCTIONS = {
        'laurentzian':"""
# Laurentzian
def fit_func(xs, A=1, gamma=1, x0=0, y0=0):
    return A/(1+(2*(xs-x0)/gamma)**2)+y0
    """,
        'gaussian':"""
# Gaussian
def fit_func(xs, A=1, sigma=1, x0=0, y0=0):
    return A*np.exp(-0.5*((xs-x0)/sigma)**2)+y0
    """,
        'cos':"""
# Cosine
def fit_func(xs, A=1, T=50, phi=0, y0=1):
    return A*np.cos(np.pi*xs/T+phi)+y0
    """,
        'linear':"""
# Linear
def fit_func(xs, m=1, y0=0):
    return m*xs + y0
    """,
    }

    def __init__(self, w, parent=None):
        super().__init__(parent=parent)
        self.w = w
        self.traces = dict()
        self.fits = dict()
        self.init_ui()
        return

    def init_ui(self):
        self.traces_list = QtWidgets.QComboBox()

        self.editor = Scintilla_Code_Editor()
        lexer = Monokai_Python_Lexer(self.editor)
        self.editor.setLexer(lexer)

        self.func_list = QtWidgets.QComboBox()
        self.func_list.addItems([func_name for func_name in sorted(self.DEFAULT_FUNCTIONS.keys())])
        def update_text(*args):
            selected_func = self.func_list.currentText()
            txt = self.EXTRA_TEXT + self.DEFAULT_FUNCTIONS[selected_func]
            self.editor.setText(txt)
        self.func_list.currentIndexChanged.connect(update_text)
        update_text()

        self.results_table = QtWidgets.QTableWidget()

        self.update_traces()
        self.fit = QtWidgets.QPushButton('Fit')
        self.remove = QtWidgets.QPushButton('Remove fit')
        self.remove_all = QtWidgets.QPushButton('Remove all fits')
        self.fit.clicked.connect(self.compile_and_fit)
        self.remove.clicked.connect(self.remove_fit)
        self.remove_all.clicked.connect(self.remove_all_fits)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.traces_list)
        layout.addWidget(self.func_list)
        layout.addWidget(self.editor)
        layout.addWidget(self.results_table)
        layout.addWidget(self.fit)
        layout.addWidget(self.remove)
        layout.addWidget(self.remove_all)
        self.setLayout(layout)
        return

    def update_results(self):
        selected_trace_name = self.traces_list.currentText()
        if selected_trace_name in self.fits:
            self.results_table.clear()
            self.results_table.setColumnCount(2)
            self.results_table.setHorizontalHeaderLabels(['param', 'value'])

            params = self.fits[selected_trace_name]['params']
            self.results_table.setRowCount(len(params))

            for i, name in enumerate(params.keys()):
                self.results_table.setItem(i,0, QtWidgets.QTableWidgetItem(name))
                self.results_table.setItem(i,1, QtWidgets.QTableWidgetItem(str(params[name])))

    def update_traces(self):
        self.traces_list.clear()
        self.traces_list.addItems([plot_name for plot_name in sorted(self.traces.keys())])
        return

    def update_fits(self):
        for trace_name, data in self.fits.items():
            xs = data['xs']
            ys = data['ys']
            curve = data['curve']
            if curve is None:
                curve = self.w.plot(pen=pg.mkPen(color=colors['yellow'], width=1), antialias=True)
                data['curve'] = curve
            curve.setData(x=xs, y=ys)
        return

    def remove_fit(self):
        selected_trace_name = self.traces_list.currentText()
        try:
            fit_data = self.fits[selected_trace_name]
            fit = fit_data['curve']
        except KeyError:
            return
        self.w.removeItem(fit)
        del self.fits[selected_trace_name]
        return

    def remove_all_fits(self):
        for name in (self.traces_list.itemText(idx) for idx in range(self.traces_list.count())):
            try:
                fit_data = self.fits[name]
                fit = fit_data['curve']
            except KeyError:
                continue
            self.w.removeItem(fit)
            del self.fits[name]
        return

    def compile_and_fit(self):
        selected_trace_name = self.traces_list.currentText()
        trace, _ = self.traces[selected_trace_name]
        plot_xs, plot_ys = trace.xData, trace.yData

        # Execute the code
        glob = globals()
        exec(self.editor.text(), glob)
        fit_func = glob['fit_func']

        # Get some default values
        sig = inspect.signature(fit_func)
        p0 = list()
        for i, p in enumerate(sig.parameters.values()):
            if i == 0:
                pass # To remove the xs parametter
            elif p.default is inspect._empty:
                p0.append(1)
            else:
                p0.append(p.default)

        # Perform the curve fitting
        popt, pcov = curve_fit(fit_func, plot_xs, plot_ys, p0=p0)
        fit_ys = fit_func(plot_xs, *popt)

        names = list(sig.parameters.keys())
        params = OrderedDict([(names[i+1], popt[i]) for i in range(len(popt))])

        self.fits[selected_trace_name] = {
            'xs': plot_xs,
            'ys': fit_ys,
            'curve': self.fits.get(selected_trace_name, dict()).get('curve'),
            'params': params,
            'pcov': pcov
        }
        self.update_fits()
        self.update_results()



#!/usr/bin/env python
import os
import logging

import pandas as pd

from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QApplication

__package__ = 'nspyre.gui'

from .widgets.splitter_widget import Splitter, SplitterOrientation
from .data_handling import load_data
from .view_manager import ViewManagerWindow

logger = logging.getLogger(__name__)


class Permanent_QFileDialog(QtWidgets.QFileDialog):
    def done(self, r):
        print('done')


class Local_json_DB(QtCore.QObject):
    """This essentially emulates a nspyre.mongo_listner.Synched_Mongo_Database by implementing
    a dfs proprety and a get_df function"""
    updated_row = QtCore.pyqtSignal(object, object)  # Emit the updated row in the format (col_name, row)
    col_added = QtCore.pyqtSignal(object)  # Emit the name of the collection which was added
    col_dropped = QtCore.pyqtSignal(object)  # Emit the name of the collection which was dropped
    db_dropped = QtCore.pyqtSignal()  # Emitted when the database is dropped

    def __init__(self, filename=None):
        super().__init__()
        self.dfs = dict()
        if not filename is None:
            self.load(filename)

    def load(self, filename):
        self.clear()
        d = load_data(filename)
        # name = d['name']
        # description = d['description']

        self.dfs = {d['spyrelet_name']:d['data']}
        self.dfs['Register'] = pd.DataFrame({'_id':[d['spyrelet_name']],'class':[d['spyrelet_class']]}).set_index('_id')
        self.col_added.emit(d['spyrelet_name'])

        for cname in d['children']:
            num = len(d['children'][cname]['data_list'])
            for i in range(num):
                cclass = d['children'][cname]['spyrelet_class']
                cdata = d['children'][cname]['data_list'][i]
                self.add_df('{}_{}'.format(cname, i), cdata, cclass)
        return d
                
    
    def add_df(self, name, df, sclass):
        reg_df = pd.DataFrame({'_id':[name], 'class':[sclass]}).set_index('_id')
        self.dfs['Register'] = self.dfs['Register'].append(reg_df)
        self.dfs[name] = df
        self.col_added.emit(name)

    def drop_df(self, name):
        self.dfs.pop(name)
        self.col_dropped.emit(name)

    def clear(self):
        for name in list(self.dfs.keys()):
            self.drop_df(name)

    def get_df(self, df_name):
        return self.dfs[df_name]


class Data_Explorer(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.db = Local_json_DB()
        layout = QtWidgets.QVBoxLayout()
        self.view_manager = ViewManagerWindow(db=self.db, react_to_drop=True)
        self.file_browser = Permanent_QFileDialog(filter = '*.json')
        
        hsplitter = Splitter(main_w=self.file_browser, side_w=self.view_manager, orientation=SplitterOrientation.vertical_right_button)
        hsplitter.setSizes([1, 400])
        hsplitter.setHandleWidth(10)
        layout.addWidget(hsplitter)
        self.setLayout(layout)

        self.file_browser.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        self.file_browser.fileSelected.connect(self.load_file)
        self.file_browser.currentChanged.connect(self.load_file)
        self.show()

    def load_file(self, filename):
        if not os.path.isfile(filename):
            return
        sname, vname = self.view_manager.get_view_name(self.view_manager.tree.currentItem())
        print('Loading {}'.format(filename))
        d = self.db.load(filename)
        if sname in self.view_manager.items:
            if vname in self.view_manager.items[sname]:
                item = self.view_manager.items[sname][vname]
            else:
                item = self.view_manager.items[sname]['__top']
        else:
            # Open the first plot in the main spyrelet
            sname = d['spyrelet_name']
            view_names = list(self.view_manager.items[sname].keys())
            view_names.sort()
            # view_names.pop('__top')
            item = self.view_manager.items[sname][view_names[1]] # Element 0 should be __top
        self.view_manager.tree.setCurrentItem(item)


if __name__ == '__main__':
    import logging
    import sys
    from PyQt5.QtCore import Qt
    from nspyre.gui import NSpyreApp
    from nspyre.misc import nspyre_init_logger

    nspyre_init_logger(logging.INFO)

    logger.info('starting Data Explorer...')
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp([sys.argv])
    data_explorer = Data_Explorer()
    sys.exit(app.exec())



"""
pytest fixtures

Author: Jacob Feder
Date: 11/12/2020
"""

###########################
# imports
###########################

# std
from pathlib import Path
import subprocess
import atexit
import time
import logging

# 3rd party
import pytest
import psutil

# nspyre
from nspyre.inserv.gateway import InservGateway

###########################
# globals
###########################

server_cfg_path = Path(__file__).parent / Path('fixtures/configs/server_test_config.yaml')
client_cfg_path = Path(__file__).parent / Path('fixtures/configs/client_test_config.yaml')

###########################
# fixtures
###########################

@pytest.fixture()
def client_config_path():
    """return the client config path"""
    logging.info('getting client config')
    return client_cfg_path.resolve()

@pytest.fixture(scope='class')
def gateway():
    """return an instrument gateway"""
    logging.info('getting gateway')
    with InservGateway(client_cfg_path) as insgw:
        yield insgw

@pytest.fixture(scope='session', autouse=True)
def setup():
    """start mongodb and the instrument server in subprocesses for use by 
    subsequent tests"""
    logging.info('test setup...')

    # search through all running processes, and only start mongo if it's not
    # already running, since it takes awhile to start up
    if not 'mongod' in [p.name() for p in psutil.process_iter()]:
        logging.info('running nspyre-mongodb')
        # start mongod in a subprocess
        mongo = subprocess.run(['nspyre-mongodb'])
        # give time for the database to start
        time.sleep(30)

    # start the instrument server
    inserv = subprocess.Popen(['nspyre-inserv', '-c', server_cfg_path, '-v', 'debug'],
                                stdin=subprocess.PIPE)

    # make sure the inserv gets killed on exit even if there's an error
    def cleanup():
        inserv.kill()
    atexit.register(cleanup)

    # ignore logging while we attempt to connect
    logging.disable(logging.CRITICAL)
    # wait until the server is online
    while True:
        try:
            with InservGateway(client_cfg_path) as insgw:
                getattr(insgw, 'tserv')
                break
        except:
            time.sleep(0.1)
    # re-enable logging
    logging.disable(logging.NOTSET)

    # now the tests run
    yield

    logging.info('test teardown')



import numpy as np
from lantz.core import Action, Driver, Feat

class FakeSigGen(Driver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._amplitude = 0.0
        self._frequency = 1e3

    @Feat(units='Hz')
    def frequency(self):
        return self._frequency
    @frequency.setter
    def frequency(self, value):
        self._frequency = value

    @Feat(units='V')
    def amplitude(self):
        return self._amplitude
    @amplitude.setter
    def amplitude(self, value):
        self._amplitude = value

    @Action()
    def calibrate(self):
        pass


import traceback
import time

from collections import OrderedDict

import json
import pandas as pd

from nspyre.misc import get_mongo_client


def get_exp_state_from_db(mongodb_addr=None, debug=False):
    c = get_mongo_client(mongodb_addr=mongodb_addr)

    state = OrderedDict()
    dbs = [dbname for dbname in c.list_database_names() if 'Instrument_Server' in dbname]
    for dbname in dbs:
        for dname in c[dbname].list_collection_names():
            dev_col = c[dbname][dname]
            state[dname] = OrderedDict()
            try:
                for param in dev_col.find({},{'_id':False, 'name':True, 'type':True, 'units':True, 'value':True}):
                    if 'type' in param and 'name' in param:
                        if param['type'] == 'feat':
                            if param['units'] is None:
                                state[dname][param['name']] = param['value']
                            else:
                                state[dname][param['name']] = str(param['value']) + " " + param['units']
                        elif param['type'] == 'dictfeat':
                            if param['units'] is None:
                                state[dname][param['name']] = param['value']
                            else:
                                state[dname][param['name']] = [str(val) + " " + param['units'] for val in param['value']]
            except:
                print("Could not save the state for device: {}".format(dname))
                if debug: traceback.print_exc()
    return state


def gen_exp_state(mode='current', mongodb_addr=None, debug=False):
    if mode == 'current':
        return get_exp_state_from_db(mongodb_addr=mongodb_addr, debug=debug)
    # elif mode =='renew':
    #     #Here will directly query all the instrument for their states
    #     raise NotImplementedError()
    #     state_dict = OrderedDict()
    #     try:
    #         m = Instrument_Manager()
    #     except:
    #         print("Could not start the instrument manager")
    #         if debug: traceback.print_exc()
    # elif mode == 'disable':
    #     return {}
    else:
        raise Exception("Invalid mode to generate the experimental state")


def save_data(spyrelet, filename, name=None, description=None, save_state_mode='current', debug=False, **kwargs):
    d = spyrelet.data.drop(['_id'], axis=1)
    data_dict = OrderedDict([
        ('name',name),
        ('description',description),
        ('spyrelet_name',spyrelet.name),
        ('spyrelet_class',"{}.{}".format(spyrelet.__class__.__module__, spyrelet.__class__.__name__)),
        ('date', time.strftime('%Y-%m-%d')),
        ('time', time.strftime('%H:%M:%S')),
        ('save_format', 'spyrelet 1.0'),
        ('experimental_state', None),
        ('data_col', list(d.columns)),
        ('data', d.to_json(orient='values')),
    ])

    try:
        child_dict = OrderedDict()
        for c_name, data_list in spyrelet.child_data.items():
            c_real_name = next(real_name for real_name, obj in spyrelet.spyrelets.items() if obj.name == c_name)#This is not ideal...
            child_dict[c_name] = OrderedDict([
                ('spyrelet_class',"{}.{}".format(getattr(spyrelet,c_real_name).__class__.__module__, getattr(spyrelet,c_real_name).__class__.__name__)),
                ('data_col', list(data_list[0].drop(['_id'], axis=1).columns)),
                ('data_list',[dl.drop(['_id'], axis=1).to_json(orient='values') for dl in data_list]),
            ])
        data_dict['children'] = child_dict
    except:
        print("Could generate child dict")
        data_dict['children'] = child_dict
        if debug: traceback.print_exc()

    #This takes care of writting the experimental state
    state_dict = dict()
    try:
        if not save_state_mode is None:
            state_dict = gen_exp_state(mode=save_state_mode, debug=debug)
    except:
        print("Could generate state dict")
        if debug: traceback.print_exc()

    #Write the file
    finally:
        data_dict['experimental_state'] = state_dict
        data_dict.update(kwargs)

        if filename is None:
            return data_dict
        else:
            with open(filename, 'w') as f:
                results = json.dump(data_dict, f, indent=4)
            return results


def load_data(filename, load_to_mongo=False, mongodb_addr=None, db_name='Spyre_Data_Loaded'):
    with open(filename, 'r') as f:
        ans = json.load(f)
    ans['data'] = pd.read_json(ans['data']).rename(columns={i:x for i,x in enumerate(ans['data_col'])})

    for c_name, c_dict in ans['children'].items():
        data_list = list()
        for d in c_dict['data_list']:
            data_list.append(pd.read_json(d).rename(columns={i:x for i,x in enumerate(c_dict['data_col'])}))
        ans['children'][c_name]['data_list'] = data_list

    if load_to_mongo:
        client = get_mongo_client(mongodb_addr)
        client.drop_database(db_name)
        db = client[db_name]
        def add_spyrelet_data(sname, sclass, data):
            reg_entry = {'_id':sname,'class':sclass}
            db['Register'].update_one({'_id':sname},{'$set':reg_entry}, upsert=True)
            db[sname].insert_many(data.to_dict(orient='records'))
        
        add_spyrelet_data(ans['spyrelet_name'], ans['spyrelet_class'], ans['data'])
        for c_name, c_dict in ans['children'].items():
            for i, d in enumerate(c_dict['data_list']):
                add_spyrelet_data(c_name+'_'+str(i), c_dict['spyrelet_class'], d)
    return ans



"""The class defining the NSpyre MainWindow.

The NSpyreMainWindow class is the main GUI window for the NSpyre software application.
It is defined by the QtWidgets.QMainWindow subclass from Qt and defines
the process for launching all the other QtWidgets.QtWidget for the other GUI windows.
i.e. Instrument Manager, View Manager, Spyrelet Launcher, etc.

From the Qt documentation:
The QMainWindow class provides a main application window, with a menu bar, dock
windows (e.g. for toolbars), and a status bar. Main windows are most often used to
provide menus, toolbars and a status bar around a large central widget, such as a
text edit, drawing canvas or QWorkspace (for MDI applications). QMainWindow is
usually subclassed since this makes it easier to encapsulate the central widget,
menus and toolbars as well as the window's state. Subclassing makes it possible to
create the slots that are called when the user clicks menu items or toolbar buttons.

  Typical usage example:

  app = app.NSpyreApp([sys.argv])
  window = NSpyreMainWindow()
  sys.exit(app.exec())

Copyright (c) 2020, Alexandre Bourassa, Michael Solomon, Jacob Feder
All rights reserved.

This work is licensed under the terms of the 3-Clause BSD license.
For a copy, see <https://opensource.org/licenses/BSD-3-Clause>.
"""
import functools
from pathlib import Path
import logging
from subprocess import Popen

from PyQt5.QtCore import QProcess
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QMainWindow, QPushButton, QVBoxLayout, QWidget

from nspyre.definitions import LOGO_PATH

from .image import ImageWidget

logger = logging.getLogger(__name__)

HERE = Path(__file__).parent


class NSpyreMainWindow(QMainWindow):
    """The Qt QtWidgets.QMainWindow object for launching NSpyre.

    This is the class you need to instantiate for starting the main
    NSpyre GUI. This defines the layout of the NSpyre launcher and
    is responsible for correctly starting the Qt Widgets for the
    additional GUI windows in NSpyre.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('NSpyre')

        # Set the main window layout to consist of vertical boxes.
        # The QVBoxLayout class lines up widgets vertically.
        layout = QVBoxLayout()

        # Add the NSpyre launcher logo as the top GUI element
        image = ImageWidget(LOGO_PATH)
        layout.addWidget(image)

        # Add a QtPushButton element for launching each additional
        # GUI window and connect a function for launching each
        # Qt Widget correctly
        inserv_manager_button = QPushButton('Instrument Manager', self)
        inserv_manager_button.setFont(QFont('Helvetica [Cronyx]', 16))
        inserv_manager_button.clicked.connect(
            functools.partial(self._launch_window, window_name='inserv_manager'))

        view_manager_button = QPushButton('View Manager', self)
        view_manager_button.setFont(QFont('Helvetica [Cronyx]', 16))
        view_manager_button.clicked.connect(
            functools.partial(self._launch_window, window_name='view_manager'))

        spyrelet_startup_button = QPushButton('Startup Spyrelet(s)', self)
        spyrelet_startup_button.setFont(QFont('Helvetica [Cronyx]', 16))
        spyrelet_startup_button.clicked.connect(
            functools.partial(self._launch_window, window_name='spyrelet_startup'))

        data_explorer_button = QPushButton('Data Explorer', self)
        data_explorer_button.setFont(QFont('Helvetica [Cronyx]', 16))
        data_explorer_button.clicked.connect(
            functools.partial(self._launch_window, window_name='data_explorer'))

        # add each button to the box layout as a widget
        layout.addWidget(inserv_manager_button)
        layout.addWidget(view_manager_button)
        layout.addWidget(spyrelet_startup_button)
        layout.addWidget(data_explorer_button)

        w = QWidget()
        w.setLayout(layout)
        self.setCentralWidget(w)
        self.show()

    def _launch_window(self, window_name=None):
        """Spawn an additional window in a new process."""

        if window_name == 'inserv_manager':
            logger.info('starting Instrument Manager...')
            Popen(['python', str(HERE.joinpath('instrument_manager.py'))])
        elif window_name == 'view_manager':
            logger.info('starting View Manager...')
            Popen(['python', str(HERE.joinpath('view_manager.py'))])
        elif window_name == 'spyrelet_startup':
            logger.info('starting Syrelet GUI window...')
            Popen(['python', str(HERE.joinpath('launcher.py'))])
        elif window_name == 'data_explorer':
            logger.info('starting Data Explorer...')
            Popen(['python', str(HERE.joinpath('data_explorer.py'))])
        else:
            raise ValueError('Incorrect input for window_name: {}'.format(window_name))



#!/usr/bin/env python
import argparse

from pathlib import Path
from subprocess import check_call, CalledProcessError
from sys import platform, exit

from nspyre.errors import OSNotSupportedError

from .mongo_listener import (
    Synched_Mongo_Database
)

__all__ = [
    'Synched_Mongo_Database'
]


THIS_DIR = Path(__file__).parent


def main():
    """Entry point for mongodb CLI"""
    # parse command-line arguments
    arg_parser = argparse.ArgumentParser(prog='nspyre-mongodb', description='Start / restart the MongoDB server')
    cmd_args = arg_parser.parse_args()
    try:
        if platform == 'linux' or platform == 'linux2' or platform == 'darwin':
            check_call(['bash', str(THIS_DIR / 'start_mongo_unix.sh')])
        elif platform == 'win32':
            check_call([str(THIS_DIR / 'start_mongo_win.bat')])
        else:
            raise OSNotSupportedError('Your OS [{}] is not supported'.\
                                        format(platform))
    except CalledProcessError:
        exit(1)

    exit(0)


if __name__ == '__main__':
    main()



"""The base class for spyrelets and associated utility functions.

The Spyrelet class is the base class for writing experimentation logic. These
can be as simple or complex as desired and need to be Leave one blank line.  The rest of this docstring should contain an
overall description of the module or program.  Optionally, it may also
contain a brief description of exported classes and functions and/or usage
examples.

  Typical usage example:

  foo = ClassFoo()
  bar = foo.FunctionBar()

Copyright (c) 2020, Alexandre Bourassa, Michael Solomon, Jacob Feder
All rights reserved.

This work is licensed under the terms of the 3-Clause BSD license.
For a copy, see <https://opensource.org/licenses/BSD-3-Clause>.
"""
from collections import OrderedDict
import copy
import inspect
import logging
import pathlib
import threading
import traceback

import numpy as np
import pandas as pd
from pint.util import infer_base_unit
from pymongo.errors import PyMongoError
from tqdm.auto import tqdm

from nspyre.config import get_config_param
from nspyre.definitions import Q_
from nspyre.errors import EntryNotFoundError, SpyreletLoadError, SpyreletRunningError, SpyreletUnloadError
from nspyre.gui import save_data
from nspyre.misc import custom_decode, custom_encode, get_mongo_client, load_class_from_file, RangeDict, deprecated

logger = logging.getLogger(__name__)

# config file key for spyrelets
CONFIG_SPYRELETS_KEY = 'spyrelets'
# config file key for the python file of the spyrelet
CONFIG_SPYRELETS_FILE_KEY = 'file'
# config file key for the class of the spyrelet
CONFIG_SPYRELETS_CLASS_KEY = 'class'
# config file key for the device aliases of a spyrelet
CONFIG_DEVS_KEY = 'device_aliases'
# config file key for the sub-spyrelets of a spyrelet
CONFIG_SUB_SPYRELETS_KEY = 'spyrelets'
# config file key for spyrelet arguments
CONFIG_SPYRELETS_ARGS_KEY = 'args'
# config file key for spyrelet keyword arguments
CONFIG_SPYRELETS_KWARGS_KEY = 'kwargs'
# record of the loaded spyrelets
_LOADED_SPYRELETS = {}


class Spyrelet:
    """The base class for constructing experiment logic.

    This is the class you need to subclass for making experiments.
    All devices used in the spyrelet must be listed in the
            REQUIRED_DEVICES dict
    All sub-spyrelets must also be listed in the REQUIRED_SPYRELETS dict
    Upon instantiation the class will check the __init__ arguments
            devices and spyrelets to make sure they satisfy these requirements
    For higher performance we will store the data internally as a list
            instead of a dataframe.  Quicker to append to a list.

    Attributes:
        REQUIRED_DEVICES: A dict with the names and associated class of the required devices.
        REQUIRED_SPYRELETS: A dict with the name and associated class of the required sub-spyrelet(s).
        PARAMS: A definition of the parameters that are used as arguments to the main/initialize/finalize functions.
                These are used both to generate a launcher GUI and to enforce units at call time.
        CONSTS: An extra dictionary, which can be defined by the user at initialization time.
    """

    REQUIRED_DEVICES = list()
    REQUIRED_SPYRELETS = dict()
    PARAMS = dict()
    CONSTS = dict()

    def __init__(self, unique_name, gateway, device_aliases={}, spyrelets={}, mongodb_addr=None, **consts):
        self.name = unique_name
        self.progress = tqdm
        self.spyrelets = spyrelets
        self.CONSTS = self.CONSTS.copy()
        self.CONSTS.update(**consts)
        self.last_kwargs = dict()

        self.mongodb_addr = mongodb_addr
        self.validate()
        
        reg_entry = {
            '_id': unique_name,
            'class': 'unused'  # "{}.{}".format(self.__class__.__module__, self.__class__.__name__),
        }

        self.client = get_mongo_client(mongodb_addr)
        self.col = self.client['Spyre_Live_Data'][unique_name]
        self.client['Spyre_Live_Data']['Register'].update_one({'_id': unique_name}, {'$set': reg_entry}, upsert=True)
        self.clear_data()
        
        # check that the server devices are accessible and add them as
        # instance variables
        for dev_alias in self.REQUIRED_DEVICES:
            try:
                # e.g. "local1/dev1"
                dev_accessor = device_aliases[dev_alias]
            except Exception as exc:
                raise SpyreletLoadError(exc, 'Spyrelet [{}] requires the '
                                             'device [{}] but it wasn\'t defined in the [{}] '
                                             'section of the config file'.format(unique_name, dev_alias, CONFIG_DEVS_KEY)) from None
            try:
                server_name, device_name = device_aliases[dev_alias].split('/')
            except:
                raise SpyreletLoadError(None, 'Spyrelet [{}] with the '
                                              'device alias [{}] has an invalid device accessor [{}].'
                                              'The accessor should be in the form '
                                              '"server_name/device_name"'.format(unique_name, dev_accessor, dev_alias))
            try:
                server = getattr(gateway, server_name)
            except:
                raise SpyreletLoadError(None, 'Spyrelet [{}] requires the '
                                              'device [{}] (alias [{}]) but the instrument '
                                              'server is unreachable'.format(unique_name, dev_accessor, dev_alias))
            try:
                device = getattr(server, device_name)
            except:
                raise SpyreletLoadError(None, 'Spyrelet [{}] requires the '
                                              'device [{}] (alias [{}]) but the instrument '
                                              'server doesn\'t contain the device'.format(unique_name, dev_accessor,
                                                                                          dev_alias))
            setattr(self, dev_alias, device)

        # check that the sub spyrelets are loaded and add them as
        # instance variables
        for sname, sclass in self.REQUIRED_SPYRELETS.items():
            if sname in spyrelets:# and isinstance(spyrelets[sname], sclass):
                setattr(self, sname, spyrelets[sname])
            else:
                raise SpyreletLoadError('Spyrelet [{}] requires the sub-spyrelet [{}] but it wasn\'t loaded'.format(unique_name, sname))

    def set_defaults(self, **params_dict):
        d = {'defaults.{}'.format(key): val for key, val in custom_encode(params_dict).items()}
        if len(d):
            return self.client['Spyre_Live_Data']['Register'].update_one({'_id': self.name}, {'$set': d}, upsert=True)

    def run(self, *args, **kwargs):
        self.progress = kwargs.pop('progress') if ('progress' in kwargs) else tqdm
        clear_data = kwargs.pop('clear_data') if ('clear_data' in kwargs) else True
        if self.progress is None: self.progress = lambda *args, **kwargs: tqdm(*args, leave=False, **kwargs)
        try:
            args, kwargs = self._enforce_args_units(*args, **kwargs)
            self._stop_flag = False
            if clear_data:
                self.clear_data()
            self.initialize(*args, **kwargs)
            self.main(*args, **kwargs)
        except SpyreletRunningError:
            print('stopping spyrelet')
            pass
        except:
            traceback.print_exc()
        finally:
            self.finalize(*args, **kwargs)

    def bg_run(self, *args, **kwargs):
        t = threading.Thread(target=lambda: self.run(*args, **kwargs))
        t.start()
        return t

    def _enforce_args_units(self, *args, **kwargs):
        """This call cross checks the given args and kwargs with those in the signature of the main and the
        the PARAMS dictionary to enforce appropriate units and use default values for missing kwargs by
        the following order of priority:
        1. kwarg given as input
        2. kwarg default given by main signature
        3. kwarg default given by PARAMS
        """
        args = list(args)
        main_sig = inspect.signature(self.main)
        main_parameters = list(main_sig.parameters.keys())

        def _enforce_units(val, param):
            """Checks if the input value has an associated unit and
            returns a Quantity object in base units; otherwise returns
            the input value.
            """
            if 'units' in param:
                if type(val) is Q_:
                    return val.to(param['units'])
                else:
                    return val * Q_(1, param['units'])
            else:
                return val

        # loop through the parameters from PARAMS
        for param_name in list(self.PARAMS.keys()):
            param_index = main_parameters.index(param_name)

            # check if current parameter is an arg or kwarg
            if param_index >= len(args):

                # check if the current kwarg was given, otherwise use the default value from the main signature,
                # and if that is empty then use the default value specified in PARAMS.
                if param_name in kwargs:
                    kwargs[param_name] = _enforce_units(kwargs[param_name], self.PARAMS[param_name])
                elif main_sig.parameters[param_name].default == inspect._empty:

                    # check if the current parameter has an associated unit and create the appropriate
                    # Quantity object if needed
                    if 'units' in self.PARAMS[param_name]:
                        kwargs[param_name] = Q_(self.PARAMS[param_name]['default'], self.PARAMS[param_name]['units'])
                    else:
                        kwargs[param_name] = self.PARAMS[param_name]['default']
                else:
                    kwargs[param_name] = _enforce_units(main_sig.parameters[param_name].default, self.PARAMS[param_name])
            else:
                args[param_index] = _enforce_units(args[param_index], self.PARAMS[param_name])
        return args, kwargs

    def main(self, *args, **kwargs):
        """This is the method that will contain the user main logic.
        Should be overwritten.
        """
        raise NotImplementedError

    # pymongo does not like the deprecated decorator, so it's not being used for now.
    # @deprecated('Consolidate code into the Spyrelet main method. This will be removed in a future version of nspyre.')
    def initialize(self, *args, **kwargs):
        """This is the method that will contain the user initialize logic.
        Should be overwritten.
        """
        pass

    def finalize(self, *args, **kwargs):
        """This is the method that will contain the user finalize logic.
        Should be overwritten. This will run even if the initialize or main
        errors out.
        """
        pass

    def call(self, spyrelet, *args, **kwargs):
        """This is the proper way to call a sub spyrelet. It will take care of \
        keeping the data and calling the proper progress bar. If use_defaults \
        is True (defaults to True) every argument needs to be passed as a \
        keyword argument (ie *args will be ignored).
        """
        keep_data = kwargs.pop('keep_data') if 'keep_data' in kwargs else True
        # use_defaults = kwargs.pop('use_defaults') if 'use_defaults' in kwargs else True
        ignore_child_progress = kwargs.pop('ignore_child_progress') if 'ignore_child_progress' in kwargs else False

        if ignore_child_progress:
            progress = lambda x: x
        else:
            progress = self.progress

        # if not use_defaults:
        spyrelet.run(*args, progress=progress, **kwargs)
        # else:
        #     launcher = SpyreletLauncher(spyrelet)
        #     launcher.run(progress=self.progress,**kwargs)
        
        if keep_data:
            if spyrelet.name not in self._child_data:
                self._child_data[spyrelet.name] = list()
            self._child_data[spyrelet.name].append(spyrelet.data)

    def stop(self):
        self._stop_flag = True
        for sname in self.REQUIRED_SPYRELETS:
            getattr(self, sname).stop()

    def clear_data(self):
        self.col.drop()
        self._data = list()
        self._child_data = dict()

    CASTING_DICT = {np.int32: int, np.float64: float}

    def acquire(self, row):
        # Cleanup row
        # Here we will keep numpy arrays as is for local copy, but change it to list for MongoDB
        if not row is None:
            restore_row = dict()
            for k, val in row.items():
                if type(val) == Q_:
                    base_unit = '*'.join('{} ** {}'.format(u, p) for u, p in infer_base_unit(val).items())
                    row[k] = row[k].to(Q_(base_unit)).m
                if type(val) == np.ndarray:
                    restore_row[k] = row[k]
                    row[k] = row[k].tolist()
                if type(val) in self.CASTING_DICT:
                    row[k] = self.CASTING_DICT[type(val)](row[k])

            self.col.insert_one(row)

            for k, val in restore_row.items():
                row[k] = val
            self._data.append(row)
        if self._stop_flag:
            raise SpyreletRunningError

    """
    The cache allows for passing information from the spyrelet to remote monitoring processes.
    This cache is meant to be temporary, so unlike data which simply accumulates, this can be overwritten asynchronously
    """

    def reg_cache_clear(self):
        self.client['Spyre_Live_Data']['Register'].update_one({'_id': self.name}, {'$set': {'cache': {}}}, upsert=True)

    def reg_cache_store(self, **kwargs):
        d = {'cache.{}'.format(key): val for key, val in custom_encode(kwargs).items()}
        if len(d):
            return self.client['Spyre_Live_Data']['Register'].update_one({'_id': self.name}, {'$set': d}, upsert=True)

    def reg_cache_get(self):
        ans = self.client['Spyre_Live_Data']['Register'].find_one({'_id': self.name}, {'_id': False, 'cache': True})
        return custom_decode(ans['cache']) if 'cache' in ans else {}

    @property
    def child_data(self):
        return self._child_data

    @property
    def data(self):
        return pd.DataFrame(self._data)

    def save(self, filename, **kwargs):
        return save_data(self, filename, **kwargs)

    def validate(self):
        # Check that the signature for main, initialize and finalize functions are the same or not defined
        init_s = inspect.signature(self.initialize)
        fina_s = inspect.signature(self.finalize)
        main_s = inspect.signature(self.main)

        for k, p in main_s.parameters.items():
            if p.kind not in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.VAR_KEYWORD]:
                raise Exception('Invalid main signature, Spyrelets do not support positional only, keyword only or var_positional arguments')

        is_default = lambda sig: list(sig.parameters.keys()) == [CONFIG_SPYRELETS_ARGS_KEY, 'kwargs'] and sig.parameters[CONFIG_SPYRELETS_ARGS_KEY].kind == inspect.Parameter.VAR_POSITIONAL and sig.parameters['kwargs'].kind == inspect.Parameter.VAR_KEYWORD

        is_equal = lambda s1, s2: list(s1.parameters.keys()) == list(s2.parameters.keys()) and all([s1.parameters[x].kind == s2.parameters[x].kind for x in s1.parameters.keys()])

        valid_init = is_default(init_s) or is_equal(init_s, main_s)
        valid_fina = is_default(fina_s) or is_equal(fina_s, main_s)

        if not (valid_fina and valid_init):
            raise Exception('Signature of initialize and finalize function must be the same as that of main for spyrelet: ' + self.name)


class SpyreletLauncher:
    def __init__(self, spyrelet):
        self.spyrelet = spyrelet
        params = spyrelet.PARAMS

        ps = inspect.signature(self.spyrelet.main).parameters
        for k, p in ps.items():
            if p.default == inspect.Parameter.empty and not k in params:
                raise Exception("Missing positional argument in PARAMS ({})".format(k))

        # First add all the params which require ordering
        def infer(default):
            return {'type': type(default)}

        self.params = OrderedDict([(k, params[k] if k in params else infer(p.default)) for k, p in ps.items() if p.kind is not inspect.Parameter.VAR_KEYWORD])
        for pname, pdescr in self.params.items():
            if not 'defaults' in pdescr and ps[pname].default != inspect._empty:
                self.params[pname].update(default=ps[pname].default)

        self.pos_or_kw_params = list(self.params.keys())
        # Add in all the other params
        for k, val in params.items():
            if not k in self.params:
                self.params[k] = val

        # Load defaults from client
        self.reg = self.spyrelet.client['Spyre_Live_Data']['Register']

    def get_defaults(self):
        ans = self.reg.find_one({'_id': self.spyrelet.name}, {'_id': False, 'defaults': True})
        return custom_decode(ans['defaults']) if 'defaults' in ans else {}

    def run(self, progress=None, **params_dict):
        self.spyrelet.set_defaults(**params_dict)
        defaults = self.get_defaults()
        defaults.update(params_dict)
        params = {k: (val.array if type(val) == RangeDict else val) for k, val in defaults.items()}
        pos_or_kw_params = [params[k] for k in self.pos_or_kw_params]
        others = {k: val for k, val in params.items() if not k in self.pos_or_kw_params}
        return self.spyrelet.run(*pos_or_kw_params, progress=progress, **others)


def load_spyrelet_class(spyrelet_name, cfg):
    """Load a spyrelet class from a file (whose location is defined in cfg)"""
    # discover spyrelet file and class
    spyrelet_path_str, _ = get_config_param(cfg, [CONFIG_SPYRELETS_KEY, spyrelet_name, CONFIG_SPYRELETS_FILE_KEY])
    spyrelet_class_name, spyrelet_cfg_path_str = get_config_param(cfg, [CONFIG_SPYRELETS_KEY, spyrelet_name, CONFIG_SPYRELETS_CLASS_KEY])
    # resolve the spyrelet file location
    # if the path isn't absolute resolve it relative to the config file
    spyrelet_path = pathlib.Path(spyrelet_path_str)
    if not spyrelet_path.is_absolute():
        spyrelet_path = pathlib.Path(spyrelet_cfg_path_str).parent / spyrelet_path
    spyrelet_path = spyrelet_path.resolve()

    if not spyrelet_path.is_file():
        raise SpyreletLoadError(None, f'spyrelet [{spyrelet_name}] file [{spyrelet_path}] doesn\'t exist')

    return load_class_from_file(spyrelet_path, spyrelet_class_name)


def load_spyrelet(spyrelet_name, gateway, sub_spyrelet=False):
    """
    recursive function that loads a spyrelet from the config
    and also loads all sub-spyrelets
    """
    cfg = gateway.config
    if spyrelet_name in _LOADED_SPYRELETS:
        if sub_spyrelet:
            return
        else:
            raise SpyreletLoadError(None, 'spyrelet [{}] is already '
                                          'defined'.format(spyrelet_name))

    # discover any sub-spyrelets
    try:
        sub_spyrelet_names, _ = get_config_param(cfg, [CONFIG_SPYRELETS_KEY, spyrelet_name, CONFIG_SUB_SPYRELETS_KEY])
    except EntryNotFoundError:
        logger.debug('spyrelet [{}] no sub-spyrelets found'.format(spyrelet_name))
        sub_spyrelet_names = {}

    # iterate through the sub-spyrelets and load them
    sub_spyrelets = {}
    for s in sub_spyrelet_names:
        if s in _LOADED_SPYRELETS:
            sub_spyrelets[s] = _LOADED_SPYRELETS[s]
        else:
            try:
                sub_spyrelets[s] = load_spyrelet(s, gateway, sub_spyrelet=True)
            except:
                raise SpyreletLoadError(None, 'spyrelet [{}] sub-spyrelet [{}] failed to load'.format(spyrelet_name, s))

    # discover and load the spyrelet class
    spyrelet_class = load_spyrelet_class(spyrelet_name, cfg)

    # discover the spyrelet devices
    try:
        dev_aliases, _ = get_config_param(cfg, [CONFIG_SPYRELETS_KEY, spyrelet_name, CONFIG_DEVS_KEY])
    except EntryNotFoundError:
        logger.debug('spyrelet [{}] no device aliases found'.format(spyrelet_name))
        dev_aliases = {}

    # discover the spyrelet arguments
    try:
        args, _ = get_config_param(cfg, [CONFIG_SPYRELETS_KEY, spyrelet_name, CONFIG_SPYRELETS_ARGS_KEY])
    except EntryNotFoundError:
        logger.debug('spyrelet [{}] no args found'.format(spyrelet_name))
        args = []
    args = custom_decode(args)

    # discover the spyrelet keyword arguments
    try:
        kwargs, _ = get_config_param(cfg, [CONFIG_SPYRELETS_KEY, spyrelet_name, CONFIG_SPYRELETS_KWARGS_KEY])
    except EntryNotFoundError:
        logger.debug('spyrelet [{}] no kwargs found'.format(spyrelet_name))
        kwargs = {}
    kwargs = custom_decode(kwargs)

    # create the spyrelet
    spyrelet = spyrelet_class(*args, spyrelet_name, gateway, **kwargs, device_aliases=dev_aliases, spyrelets=sub_spyrelets)
    _LOADED_SPYRELETS[spyrelet_name] = spyrelet
    logger.info('loaded spyrelet [{}]'.format(spyrelet_name))

    return spyrelet


def load_all_spyrelets(gateway):
    """Load all of the spyrelets from the configuration file.

    Instantiates the spyrelet classes retrieved from the given classes specified in the active
    configuration file. An InservGateway instance must be passed for handling the connection to
    the devices specificed in each spyrelet.

    Args:
        manager: An InservGateway instance connected to a running InstrumentServer.
        filepath: Optional; An absolute filepath to a configuration file. If a filepath is specified,
            the spyrelets will be loaded from this file. Otherwise, the default behavior is to load
            from the active configuration file.

    Returns:
        A dict mapping keys of spyrelet names to the corresponding instance of that spyrelet class.

    Raises:
        SpyreletLoadError: An error occurred because one (or more) spyrelet(s) is(are) already loaded.
    """
    # spyrelet parameters to parse
    spyrelet_configs, _ = copy.copy(get_config_param(gateway.config, [CONFIG_SPYRELETS_KEY]))
    # check to see if any spyrelets are loaded
    if _LOADED_SPYRELETS:
        raise SpyreletLoadError(None, 'the following spyrelets were already loaded so nothing was done: {}'.format(_LOADED_SPYRELETS))

    # parse the spyrelets, loading them as we go until there
    # are none left
    while bool(spyrelet_configs):
        spyrelet_name = next(iter(spyrelet_configs))
        load_spyrelet(spyrelet_name, gateway)
        # remove this spyrelet from the list of spyrelets to be loaded
        del spyrelet_configs[spyrelet_name]
    return _LOADED_SPYRELETS


def unload_spyrelet(name, client=None):
    """function to unload a current spyrelet"""
    if name not in _LOADED_SPYRELETS:
        raise SpyreletUnloadError(None, 'the spyrelet {} does not exist'.format(name))
    if client is None:
        client = get_mongo_client()
    try:
        client['Spyre_Live_Data'][name].drop()
        client['Spyre_Live_Data']['Register'].delete_one({'_id': name})
    except PyMongoError as error:
        raise SpyreletUnloadError(error, 'error attempting to remove spyrelet: {} from the MongoDB server'.format(name))
    _LOADED_SPYRELETS.pop(name)


def unload_all_spyrelets(except_list=[], client=None):
    """function to unload all current spyrelet(s)"""
    if client is None: client = get_mongo_client()
    all_in_reg = [x['_id'] for x in client['Spyre_Live_Data']['Register'].find({}, {})]
    for name in all_in_reg:
        if not name in except_list:
            unload_spyrelet(name, client=client)


def reload_spyrelet(spyrelet_names, gateway, client=None):
    """function to reload current spyrelet(s)"""
    # if only one name is passed, convert it to list for handling
    if type(spyrelet_names) is not list:
        spyrelet_names = [spyrelet_names]

    if client is None:
        client = get_mongo_client()

    for name in spyrelet_names:
        unload_spyrelet(name, client=client)
        load_spyrelet(name, gateway)


def reload_all_spyrelets(gateway, except_list=[], client=None):
    """function to reload all current spyrelet(s)"""
    if client is None:
        client = get_mongo_client()

    # get the symmetric difference
    spyrelets = list(set(_LOADED_SPYRELETS.keys()) ^ set(except_list))

    unload_all_spyrelets(except_list, client)
    if except_list:
        for name in spyrelets:
            load_spyrelet(name, gateway)
    else:
        load_all_spyrelets(gateway)


@property
def LOADED_SPYRELETS():
    """function to return all loaded spyrelets"""
    return _LOADED_SPYRELETS



import time
import traceback

import pandas as pd

from PyQt5 import QtCore

from nspyre.misc import get_mongo_client


class DropEvent():
    """Represents a drop of a collection in a certain database"""
    def __init__(self, db, col):
        self.db, self.col = db, col


def modify_df(df, change):
    # print(change)
    if change['operationType'] == 'drop':
        return DropEvent(change['ns']['db'], change['ns']['coll']), None
    key  = change['documentKey']['_id']
    if change['operationType'] == 'update':
        for k, val in change['updateDescription']['updatedFields'].items():
            ks = k.split('.')
            if len(ks) == 1:
                df.loc[key, k] = val
            elif len(ks) == 2:
                # TODO: Figure out a more reliable way of doing this
                
                if ks[1].isdigit():
                    # Assume an array here... Will see if we can get away with this
                    df.loc[key,ks[0]][int(ks[1])] = val
                else:
                    df.loc[key,ks[0]][ks[1]] = val
                    
            else:
                raise NotImplementedError('Cannot use a dept of more then 2 in the documents')
    elif change['operationType'] == 'insert':
        doc = change['fullDocument']
        _id = doc.pop('_id')
        s = pd.Series(doc, name=_id)
        df = df.append(s)

    else:
        raise NotImplementedError('Cannot modify df with operationType: {}'.format(change['operationType']))
    return df, df.loc[key]


class Mongo_Listenner(QtCore.QThread):
    """Qt Thread which monitors for changes to qither a collection or a database and emits a signal when something happens"""
    updated = QtCore.pyqtSignal(object)
    def __init__(self, db_name, col_name=None, mongodb_addr=None):
        super().__init__()
        self.db_name = db_name
        self.col_name = col_name
        self.mongodb_addr = mongodb_addr
        self.exit_flag = False

    def run(self):
        self.exit_flag = False
        # Connect
        client = get_mongo_client(self.mongodb_addr)
        mongo_obj = client[self.db_name]
        if not self.col_name is None:
            mongo_obj = mongo_obj[self.col_name]

        with mongo_obj.watch() as stream:
            while stream.alive:
                doc = stream.try_next()
                if doc is not None:
                    self.updated.emit(doc)
                if self.exit_flag:
                    return
        if not self.exit_flag:
            self.run() # This takes care of the invalidate event which stops the change_stream cursor


# unused in the codebase (should this be removed?)
class Synched_Mongo_Collection(QtCore.QObject):
    updated_row = QtCore.pyqtSignal(object) # Emit the updated row
    # mutex = QtCore.QMutex()
    def __init__(self, db_name, col_name, mongodb_addr=None):
        super().__init__()
        self.watcher = Mongo_Listenner(db_name, col_name=col_name, mongodb_addr=mongodb_addr)
        
        self.col = get_mongo_client(mongodb_addr)[db_name][col_name]
        self.refresh_all()

        self.watcher.start()
        self.watcher.updated.connect(self._update_df)
    
    def refresh_all(self):
        col = list(self.col.find())
        if col == []:
            self.df = None
        else:
            self.df = pd.DataFrame(col)
            self.df.set_index('_id', inplace=True)

    def get_df(self):
        # self.mutex.lock()
        return self.df

    @QtCore.pyqtSlot(object)
    def _update_df(self, change):
        if self.db is None:
            self.refresh_all()
        # print(change)
        try:
            self.df, row = modify_df(self.df, change)
            self.updated_row.emit(row)
        except:
            traceback.print_exc()
            print('Refreshing the entire collection')
            self.refresh_all()
        
        # self.refresh_all() #I will make this a little more efficient later on

    def __del__(self):
        self.watcher.exit_flag = True


class Synched_Mongo_Database(QtCore.QObject):
    updated_row = QtCore.pyqtSignal(object, object) # Emit the updated row in the format (col_name, row)
    col_added = QtCore.pyqtSignal(object) # Emit the name of the collection which was added
    col_dropped = QtCore.pyqtSignal(object) # Emit the name of the collection which was dropped
    db_dropped = QtCore.pyqtSignal() #Emitted when the database is dropped

    def __init__(self, db_name, mongodb_addr=None):
        super().__init__()
        self.watcher = Mongo_Listenner(db_name, col_name=None, mongodb_addr=mongodb_addr)
        
        self.db = get_mongo_client(mongodb_addr)[db_name]
        self.refresh_all()

        self.watcher.start()
        self.watcher.updated.connect(self._update)
    
    def refresh_all(self):
        self.dfs = dict()
        for col in self.db.list_collection_names():
            col_data = list(self.db[col].find())
            if not col_data == []:
                self.dfs[col] = pd.DataFrame(col_data)
                self.dfs[col].set_index('_id', inplace=True)

    def get_df(self, col_name, timeout=0.1):
        try:
            if not self.dfs[col_name] is None:
                return self.dfs[col_name]
        finally:
            t = time.time()
            while time.time()-t<timeout:
                if col_name in self.dfs:
                    return self.dfs[col_name]


    @QtCore.pyqtSlot(object)
    def _update(self, change):
        # print(change)
        try:
            if change['operationType'] == 'dropDatabase':
                self.dfs = dict()
                self.db_dropped.emit()
                return
            elif change['operationType'] == 'invalidate':
                return
            col = change['ns']['coll']
            if col in self.dfs:
                df, row = modify_df(self.dfs[col], change)
                if isinstance(df, DropEvent):
                    self.dfs.pop(col)
                    self.col_dropped.emit(col)
                    return
                
                self.dfs[col] = df
                self.updated_row.emit(col, row)
                
            else:
                doc = change['fullDocument']
                row = pd.Series(doc)
                self.dfs[col] = pd.DataFrame([row])
                self.dfs[col].set_index('_id', inplace=True)
                self.col_added.emit(col)
                self.updated_row.emit(col, row)
        except:
            traceback.print_exc()
            print('Refreshing the entire database')
            self.refresh_all()
        # self.refresh_all() #I will make this a little more efficient later on

    def __del__(self):
        self.watcher.exit_flag = True



#!/usr/bin/env python
import logging
import pathlib

import yaml

from nspyre.definitions import CLIENT_META_CONFIG_PATH
from nspyre.errors import ConfigurationError, EntryNotFoundError


logger = logging.getLogger(__name__)

META_CONFIG_FILES_ENTRY = 'config_files'
META_CONFIG_ENABLED_IDX = 'enabled'


# A meta-config.yaml file contains a single entry with key
# META_CONFIG_FILES_ENTRY and value = a list of all the config files that
# should be read
def load_raw_config(filepath):
    """Return a config file dictionary loaded from a YAML file"""
    with open(filepath, 'r') as f:
        conf = yaml.safe_load(f)
    return conf


def meta_config_add(meta_config_file, files):
    """Add config files to the meta-config"""
    meta_config = load_raw_config(meta_config_file)
    config_list = meta_config[META_CONFIG_FILES_ENTRY]
    new_files = []
    for f in files:
        f_path = pathlib.Path(f).resolve()
        if not f_path.is_file():
            raise FileNotFoundError('file [{}] not found'.format(f_path))
        if str(f_path) in config_list:
            raise ConfigurationError(
                None, 'the config file {} is already available'.format(f)
            )
        new_files.append(str(f_path))
    meta_config[META_CONFIG_FILES_ENTRY] = config_list + new_files
    write_config(meta_config, meta_config_file)


def meta_config_remove(meta_config_file, files):
    """Remove config files from the meta-config"""
    meta_config = load_raw_config(meta_config_file)
    enabled_idx = meta_config[META_CONFIG_ENABLED_IDX]
    config_list = meta_config[META_CONFIG_FILES_ENTRY]
    # list of indicies to remove from config_list
    pop_list = []
    # go through the list of config files / indicies and generate
    # a list of indicies to remove
    for c in files:
        try:
            idx = int(c)
        except ValueError:
            # the user passed the config name as a string, so we should first
            # find its index
            try:
                idx = meta_config[META_CONFIG_FILES_ENTRY].index(c)
            except ValueError as exc:
                raise ConfigurationError(
                    exc,
                    'config file [{}] was not found in the meta-config - check that the file path shown using --list-configs matches given input'.format(
                        c
                    ),
                ) from None
        pop_list.append(idx)

    pop_list.sort(reverse=True)
    new_enabled_idx = enabled_idx
    for idx in pop_list:
        try:
            config_list.pop(idx)
            # if an item before the enabled one was deleted we have to decrement
            # the enabled index
            if idx < enabled_idx:
                new_enabled_idx -= 1
        except IndexError as exc:
            raise ConfigurationError(
                exc,
                'tried to remove config file index [{}] that was out of range'.format(
                    idx
                ),
            )

    # if the user removed the currently enabled config
    if enabled_idx in pop_list:
        new_enabled_idx = 0

    meta_config[META_CONFIG_FILES_ENTRY] = config_list
    meta_config[META_CONFIG_ENABLED_IDX] = new_enabled_idx
    write_config(meta_config, meta_config_file)


def meta_config_files(meta_config_file):
    """Return the paths of the config files in the meta-config"""
    meta_config = load_raw_config(meta_config_file)
    config_list = meta_config[META_CONFIG_FILES_ENTRY]
    return config_list


def meta_config_enabled_idx(meta_config_file):
    """Return the index of the enabled config"""
    meta_config = load_raw_config(meta_config_file)
    return meta_config[META_CONFIG_ENABLED_IDX]


def meta_config_set_enabled_idx(meta_config_file, idx_or_str):
    """Set the index of the enabled config"""
    meta_config = load_raw_config(meta_config_file)
    try:
        idx = int(idx_or_str)
    except ValueError:
        # the user passed the config name as a string, so we should first
        # find its index
        try:
            idx = meta_config[META_CONFIG_FILES_ENTRY].index(idx_or_str)
        except ValueError as exc:
            raise ConfigurationError(
                exc,
                'config file [{}] was not found in the meta-config - check that the file path shown using --list-configs matches given input'.format(
                    idx_or_str
                ),
            ) from None
    meta_config[META_CONFIG_ENABLED_IDX] = idx
    write_config(meta_config, meta_config_file)


def load_meta_config(meta_config_path=None):
    """Takes a 'meta' config file and returns the file path of the activated
    config file"""

    # TODO this logic should be removed and meta_config_path always passed in
    if not meta_config_path:
        meta_config_path = CLIENT_META_CONFIG_PATH

    # load the meta config
    meta_config = load_raw_config(meta_config_path)
    # get the index of the enabled config file
    enabled_idx = meta_config[META_CONFIG_ENABLED_IDX]

    # get the config file path
    config_files = meta_config[META_CONFIG_FILES_ENTRY]
    if not config_files:
        raise ConfigurationError(
            None,
            'no configuration files exist - use nspyre-config --add-config to add files',
        ) from None
    cfg_path = pathlib.Path(config_files[enabled_idx])

    # resolve relative paths
    if not cfg_path.is_absolute():
        cfg_path = meta_config_path.parent / cfg_path
    cfg_path = cfg_path.resolve()

    return cfg_path


def load_config(cfg_path):
    """Load a config file and return the configuration as a dictionary"""
    try:
        cfg_dict = {str(cfg_path): load_raw_config(cfg_path)}
    except FileNotFoundError as error:
        raise ConfigurationError(
            error, 'configuration file [{}] doesn\'t exist'.format(cfg_path)
        ) from None
    return cfg_dict


def write_config(config_dict, filepath):
    """Write a dictionary to a YAML file"""
    # open the file and write it's config dictionary
    with open(filepath, 'w') as file:
        yaml.dump(config_dict, file, default_flow_style=False)


def get_config_param(config_dict, path):
    """Navigate a YAML-loaded config file and return a particular parameter
    given by 'path'. If multiple config files contain the first element of
    'path', this will attempt to navigate the the first config file it finds
    that contains the first element of 'path'."""
    first_elem = path[0]
    for conf in config_dict:
        # first find the config file containing the first path element
        loc = config_dict[conf]
        if first_elem in loc:
            # now descend into the config dictionary, following the keys
            # one-by-one in path
            for p in path:
                try:
                    loc = loc[p]
                except KeyError:
                    raise EntryNotFoundError(path) from None
            return loc, conf
    # if we reach this point the first path element wasn't found in any
    # config file entries
    raise EntryNotFoundError(path) from None



from lantz.core import Driver, Feat

class FakePellicle(Driver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._opened = False

    @Feat(values={True, False})
    def opened(self):
        return self._opened

    @opened.setter
    def opened(self, value):
        self._opened = value



#!/usr/bin/env python
"""This module serves a shell prompt allowing the user runtime control of
the instrument server
"""
import argparse
import cmd
import pathlib
import pdb
import logging
import signal

import pyvisa

from nspyre.config import load_meta_config
from nspyre.definitions import SERVER_META_CONFIG_PATH
from nspyre.errors import InstrumentServerError
from nspyre.misc import nspyre_init_logger

# InstrumentServer must be imported before InservGateway; otherwise the gateway has a weird bug
# on first initialization as seen in the Instrument Manager feat/dictfeat values (goes away
# after refreshing the values once).
from .inserv import InstrumentServer
from .gateway import InservGateway

__all__ = [
    'InservGateway',
    'InstrumentServer'
]

logger = logging.getLogger(__name__)


class InservCmdPrompt(cmd.Cmd):
    """Instrument Server shell prompt processor"""
    def __init__(self, inserv):
        super().__init__()
        self.inserv = inserv

    def emptyline(self):
        """When no command is entered"""
        pass

    def do_list(self, arg_string):
        """List all the available devices"""
        if arg_string:
            print('Expected 0 args')
            return
        for d in self.inserv._devs.keys():
            print(d)

    def do_config(self, arg_string):
        """Reload the server config files"""
        if arg_string:
            print('Expected 0 args')
            return
        # attempt to reload the config files
        try:
            self.inserv.update_config(config_file=args[0] if arg_string else None)
        except Exception as exc:
            logger.exception(exc)
            print('Failed to reload config files')
            return

    def do_dev(self, arg_string):
        """Restart the connection with a device\n<string> the device name"""
        args = arg_string.split(' ')
        if not arg_string or len(args) > 1:
            print('Expected 1 arg: device name')
            return
        dev_name = args[0]
        try:
            self.inserv.reload_device(dev_name)
        except Exception as exc:
            logger.exception(exc)
            print('Failed to reload device [{}]'.format(dev_name))
            return

    def do_dev_all(self, arg_string):
        """Restart the connection with all devices"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.reload_devices()
        except Exception as exc:
            logger.exception(exc)
            print('Failed to reload all devices')
            return

    def do_restart(self, arg_string):
        """Restart the server AND reload the config file and all devices"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.restart()
        except Exception as exc:
            logger.exception(exc)
            print('Failed to restart')
            return

    def do_server_restart(self, arg_string):
        """Restart the rpyc server"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.reload_server()
        except Exception as exc:
            logger.exception(exc)
            print('Failed to restart server')
            return

    def do_server_stop(self, arg_string):
        """Stop the rpyc server"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.stop_server()
        except Exception as exc:
            logger.exception(exc)
            print('Failed to stop server')
            return

    def do_server_start(self, arg_string):
        """Start the rpyc server"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.start_server()
        except Exception as exc:
            logger.exception(exc)
            print('Failed to start server')
            return

    def do_debug(self, arg_string):
        """Drop into the debugging console"""
        if arg_string:
            print('Expected 0 args')
            return
        pdb.set_trace()
        
    def do_quit(self, arg_string):
        """Quit the program"""
        if arg_string:
            print('Expected 0 args')
            return
        logger.info('exiting...')
        # close all open resources
        self.inserv.stop_server()
        for dev_name in list(self.inserv._devs):
            self.inserv.del_device(dev_name)
        pyvisa.ResourceManager().close()
        raise SystemExit

def main():
    """Entry point for instrument server CLI"""

    # parse command-line arguments
    arg_parser = argparse.ArgumentParser(prog='nspyre-inserv',
                            description='Run an nspyre instrument server')
    arg_parser.add_argument('-c', '--config',
                            default=None,
                            help='use the provided config file') 
    arg_parser.add_argument('-l', '--log',
                            default=None,
                            help='log to the provided file / directory')
    arg_parser.add_argument('-q', '--quiet',
                            action='store_true',
                            help='disable logging')
    arg_parser.add_argument('-v', '--verbosity',
                            default='info',
                            help='the verbosity of logging to stdout - options are: '
                            'debug, info, warning, error')
    cmd_args = arg_parser.parse_args()

    # configure server logging behavior
    if not cmd_args.quiet:
        if cmd_args.verbosity.lower() == 'debug':
            log_level = logging.DEBUG
        elif cmd_args.verbosity.lower() == 'info':
            log_level = logging.INFO
        elif cmd_args.verbosity.lower() == 'warning':
            log_level = logging.WARNING
        elif cmd_args.verbosity.lower() == 'error':
            log_level = logging.ERROR
        else:
            raise InstrumentServerError('didn\'t recognize logging level [{}]'.\
                                        format(cmd_args.verbosity)) from None
        if cmd_args.log:
            nspyre_init_logger(log_level, log_path=pathlib.Path(cmd_args.log),
                                        log_path_level=logging.DEBUG,
                                        prefix='inserv',
                                        file_size=100e6)
        else:
            # the user asked for no log file
            nspyre_init_logger(log_level)

    # get the config file
    if cmd_args.config:
        config_path = cmd_args.config
    else:
        config_path = load_meta_config(SERVER_META_CONFIG_PATH)

    # init and start RPyC server
    logger.info('starting instrument server...')
    inserv = InstrumentServer(config_path)

    # properly stop the server when a kill signal is received
    def stop_server(signum, frame):
        inserv.stop_server()
        raise SystemExit
    signal.signal(signal.SIGINT, stop_server)
    signal.signal(signal.SIGTERM, stop_server)

    # start the shell prompt event loop
    cmd_prompt = InservCmdPrompt(inserv)
    cmd_prompt.prompt = 'inserv > '
    cmd_prompt.cmdloop('instrument server started...')


if __name__ == '__main__':
    main()



"""The class defining the NSpyre Instrument Manager MainWindow.

The InstrumentManagerWindow class is the main GUI window for viewing the live
state and settings of the hardware devices connected to an NSpyre InstrumentServer.
It is defined by the QtWidgets.QMainWindow subclass from Qt and consists of a
QTreeWidget and subsequent QTreeWidgetItem(s) for displaying the attributes of
each device located on each connected Instrument Server. It is also responsible
for creating and connecting all the callback functions to the device drivers'
PySignal.Signal(s) for updating it's own GUI QtWidgets.QtWidget(s) in real-time.
i.e. QComboBox, QLineEdit, SpinBox, etc.

From the Qt documentation:
The QMainWindow class provides a main application window, with a menu bar, dock
windows (e.g. for toolbars), and a status bar. Main windows are most often used to
provide menus, toolbars and a status bar around a large central widget, such as a
text edit, drawing canvas or QWorkspace (for MDI applications). QMainWindow is
usually subclassed since this makes it easier to encapsulate the central widget,
menus and toolbars as well as the window's state. Subclassing makes it possible to
create the slots that are called when the user clicks menu items or toolbar buttons.

  Typical usage example:

  app = app.NSpyreApp([sys.argv])
  pyqtgraph._connectCleanup()
  with ..inserv.gateway.InservGateway() as isg:
      window = main_window.InstrumentManagerWindow(isg)
      sys.exit(app.exec())

Copyright (c) 2020, Alexandre Bourassa, Michael Solomon, Jacob Feder
All rights reserved.

This work is licensed under the terms of the 3-Clause BSD license.
For a copy, see <https://opensource.org/licenses/BSD-3-Clause>.
"""
import functools
import inspect
import logging

from pimpmyclass.helpers import DictPropertyNameKey
from PyQt5.QtCore import QEvent, QObject, QSize, QThread, QTimer, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QColor, QCursor, QFont
from PyQt5.QtWidgets import (QApplication, QComboBox, QHBoxLayout, QHeaderView, QLabel, QLineEdit, QMainWindow,
                             QPushButton, QToolTip, QTreeWidget, QTreeWidgetItem, QVBoxLayout, QWidget)
from pyqtgraph import _connectCleanup as pyqtgraph_connectCleanup
from pyqtgraph import SpinBox as pyqtgraph_SpinBox, ValueLabel as pyqtgraph_ValueLabel
from pint.util import infer_base_unit

from nspyre.config import load_meta_config
from nspyre.definitions import Q_, CLIENT_META_CONFIG_PATH
from nspyre.errors import InstrumentManagerError
from nspyre.inserv import InservGateway

__all__ = []

logger = logging.getLogger(__name__)

ROW_QSIZE = QSize(79, 23)
UPDATE_HEX_COLOR = '#1CE33D'
UPDATE_DISPLAY_TIME = 350
REFRESH_PROGRESS_HEX_COLOR = '#2E5E9B'


def disable_widget_scroll_wheel_event(control: QWidget) -> None:
    """Convenience function to prevent a scroll wheel event from affecting a widget unless:
    1. The widget has focus
    2. The cursor is on the widget
    """
    control.setFocusPolicy(Qt.StrongFocus)
    control.installEventFilter(MouseWheelWidgetAdjustmentGuard(control))


class MouseWheelWidgetAdjustmentGuard(QObject):
    """This QObject class contains a Qt eventFilter method to ignore mouse scroll
     wheel inputs. This is useful to apply on widgets for which:
     a) you don't want the scroll wheel to change values ever; or
     b) the widget is inside a QAbstractScrollArea and preventing the scroll wheel from panning
     the area correctly (for this the FocusPolicy must additionally be changed to StrongFocus).
    """
    def __init__(self, parent: QObject):
        super().__init__(parent)

    def eventFilter(self, qobject: QObject, event: QEvent) -> bool:
        widget: QWidget = qobject
        if event.type() == QEvent.Wheel and not widget.hasFocus():
            event.ignore()
            return True
        return super().eventFilter(qobject, event)


class RefreshValuesThread(QThread):
    """This QThread class contains a run method for use with the InstrumentManagerWindow 'refresh all' button to update
    all feat/dictfeat values in the GUI. It does the following:
    1. update the progress bar displayed while the QThread is running
    2. call each feat/dictfeat of every driver on every server to get an updated value
    3. emit a pyqtSignal with the feat/dictfeat's tuple index in the parent GUI
       getattr_func mapping list and its updated value
    """

    refresh_progress = pyqtSignal(str, float)
    value_update = pyqtSignal(int, object)

    def __init__(self, parent: QObject):
        super().__init__(parent)
        self.mapping = parent.device_to_getattr_func_mapping

    def run(self) -> None:
        num_attributes = len(self.mapping)
        for i, mapping in enumerate(self.mapping):
            percent = float('%.5f' % (i / num_attributes))
            self.refresh_progress.emit('Please Wait... ({}/{})'.format(i, num_attributes), percent)

            # attempt to first unpack each tuple as if it is for a dictfeat, then handle for a feat
            try:
                feat_name, device, dictfeat_key = mapping[1:]
                value = getattr(device, feat_name)[dictfeat_key]
            except ValueError:
                feat_name, device = mapping[1:]
                value = getattr(device, feat_name)
            self.value_update.emit(i, value)


class pyqt_LineEdit(QLineEdit):
    """This QLineEdit class modifies the existing QLineEdit with a pyqtgraph ValueLabel so it can handle SI units in the
    same way as other pyqtgraph widgets because pyqtgraph does not have its own QLineEdit implementation.
    """

    def __init__(self, suffix, siPrefix, contents: str = None, parent: QWidget = None):
        super().__init__((contents, parent) if contents else parent)
        self.valuelabel = pyqtgraph_ValueLabel(suffix=suffix, siPrefix=siPrefix)
        self.setReadOnly(True)

    def setValue(self, value):
        self.valuelabel.setValue(value)
        self.setText(self.valuelabel.generateText())


class InstrumentManagerWindow(QMainWindow):
    """Creates a GUI interface for controlling instrument attributes
    based on a dropdown / tree structure
    the top level is the instrument servers
    the next level is the devices
    the bottom level is attributes of the devices.

     # tree of dictionaries that contains all of the instrument
     # manager GUI elements
     # the top level of the dictionary is the instrument servers
     # the next level is the devices
     # the bottom level is attributes of the devices

     e.g.               ----------self.gui----------
                       /                            \
                   server1                        server2
                  /       \                      /       \
          sig-gen1         scope1        sig-gen2         laser
          /      \        /      \      /        \       /     \
       freq     ampl    trig    din[] freq      ampl  lambda  power
     """

    def __init__(self, gateway, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('NSpyre Instrument Manager')
        # connection to the instrument servers
        self.gateway = gateway

        # set main GUI layout
        self.tree = QTreeWidget()
        self.tree.setMouseTracking(True)
        self.tree.entered.connect(self.handleItemEntered)
        self.tree.setFont(QFont('Helvetica [Cronyx]', 14))
        self.tree.setColumnCount(2)
        self.tree.setMinimumHeight(self.tree.height())
        # set all rows to have the same height.
        # this provides performance improvements to the rendering time.
        # the actual height set is determined from the first QTreeWidgetItem given.
        self.tree.setUniformRowHeights(True)

        # configure the QTreeWidget Header
        header = self.tree.header()
        header.setHidden(True)
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        # need to set to False to correctly calculate minimum width values
        header.setStretchLastSection(False)

        # add 'refresh all' button for manually triggering a driver feat update
        self.device_to_getattr_func_mapping = list()
        self.refresh_all_button = QPushButton('Refresh All Values', self)
        self.refresh_button_color = self.refresh_all_button.palette().button().color().name()
        self.refresh_all_button.setProperty('class', 'refresh_all')
        self.refresh_all_button.clicked.connect(self._refresh_all)

        # generate gui elements and set minimum window dimensions before displaying
        self._create_widgets()
        self.tree.expandAll()
        # Adding 2pt of padding to the margin to remove horizontal scroll bar and 14pt of padding for the vertical scroll bar
        self.tree.setMinimumWidth(self.tree.columnWidth(0) + self.tree.columnWidth(1) + 2 + 14)
        self.tree.collapseAll()
        # start GUI with the servers expanded
        for i in range(self.tree.topLevelItemCount()):
            self.tree.topLevelItem(i).setExpanded(True)
        # need to set this last, otherwise minimum width values are not calculated correctly
        header.setStretchLastSection(True)

        # Set the main window layout to consist of vertical boxes.
        # The QVBoxLayout class lines up widgets vertically.
        layout = QVBoxLayout()
        layout.addWidget(self.tree)
        layout.addWidget(self.refresh_all_button)
        w = QWidget()
        w.setLayout(layout)
        self.setCentralWidget(w)
        layout.setContentsMargins(0, 0, 0, layout.getContentsMargins()[3])
        self.show()

    def handleItemEntered(self, index):
        """Manual cursor tracking so that QToolTips disappear after moving cursor off corresponding QLabel."""
        if index.isValid() and index.data(Qt.ToolTipRole):
            QToolTip.showText(QCursor.pos(), index.data(Qt.ToolTipRole),
                              self.tree.viewport(), self.tree.visualRect(index))

    @pyqtSlot(str, float)
    def update_refresh_all_button(self, string: str, percent: float) -> None:
        """Slot for receiving Signals with QThread processing progress to update GUI progress bar."""
        self.refresh_all_button.setText(string)
        stylesheet = ('min-height: 34px;'  # add 2px to accord for no border
                      'border: none;'
                      'border-radius: 4px;'
                      'background-color: qlineargradient('
                      'spread: pad, x1: 0, y1: 0, x2: 1, y2: 0,'
                      'stop: 1 {background_color},'
                      'stop: {percent} {progress_color},'
                      'stop: {delta} {background_color});'.format(
                            background_color=self.refresh_button_color,
                            percent=percent,
                            delta=percent + 0.0000001,
                            progress_color=REFRESH_PROGRESS_HEX_COLOR)
                      )
        self.refresh_all_button.setStyleSheet(stylesheet)

    @pyqtSlot(int, object)
    def update_treewidgetitem_widget(self, index, value) -> None:
        """Slot for receiving Signals with updated fict/dictfeat values to update the corresponding GUI element."""
        self.device_to_getattr_func_mapping[index][0](value, None)

    def _refresh_all_finished(self):
        """Cleanup code for 'refresh all' button once the spawned QThread has finished processing."""
        self.refresh_all_button.setText('Refresh All Values')
        self.refresh_all_button.setStyleSheet('background-color: {};'.format(self.refresh_button_color))
        for child in self.tree.children()[0].children():
            child.setAttribute(Qt.WA_TransparentForMouseEvents, False)
        self.refresh_all_button.setEnabled(True)

    def _refresh_all(self):
        """Spawn a QThread to get updated values for each feat/dictfeat when the 'refresh all' button is clicked."""
        self.refresh_all_button.setEnabled(False)
        for child in self.tree.children()[0].children():
            child.setAttribute(Qt.WA_TransparentForMouseEvents, True)
        refresh_thread = RefreshValuesThread(self)
        refresh_thread.refresh_progress.connect(self.update_refresh_all_button)
        refresh_thread.value_update.connect(self.update_treewidgetitem_widget)
        refresh_thread.finished.connect(self._refresh_all_finished)
        refresh_thread.start()

    def _create_widgets(self):
        """Iterate over the available servers and devices, collect their
        attributes that can be modified by the instrument manager GUI, then
        populate the self.gui
        """

        # iterate over servers
        for server_name, server in self.gateway.servers().items():
            server_tree = QTreeWidgetItem(self.tree, [server_name, ''])
            server_tree.setExpanded(True)
            server_tree.setFont(0, QFont('Helvetica [Cronyx]', 15))
            # set size hint as the first server is the first QTreeWidgetItem to be added to the tree;
            # needed to appropriately set the uniform row height.
            server_tree.setSizeHint(0, ROW_QSIZE)

            # iterate over devices
            for device_name, device in server.root._devs.items():
                device_tree = QTreeWidgetItem(server_tree, [device_name, ''])
                device_tree.setExpanded(True)
                device_tree.setFont(0, QFont('Helvetica [Cronyx]', 15))

                try:
                    # handle feats
                    # This also returns dictfeats!
                    for feat_name, feat in device._lantz_feats.items():
                        # filter out any dictfeats
                        if isinstance(feat_name, DictPropertyNameKey):
                            continue
                        feat_widget, feat_getattr_func = self._generate_feat_widget(feat, feat_name, device)
                        self.device_to_getattr_func_mapping.append((feat_getattr_func, feat_name, device))
                        # we have to use a partial here because PySignal and RPyC don't
                        # play nicely if you .connect() a lambda or other function / method
                        # to PySignal
                        feat_getattr_func.__name__ = 'InstrumentManager_getattr_func'
                        getattr_partial = functools.partial(feat_getattr_func)
                        getattr_partial = functools.update_wrapper(getattr_partial, feat_getattr_func)
                        # register the feat_getattr_func() to be called when the feat changes using
                        # pimpmyclass "ObservableProperty" mixin
                        getattr(device, feat_name + '_changed').connect(getattr_partial)

                        # set formatting and add docstring information as a tooltip
                        feat_widget.setFont(QFont('Helvetica [Cronyx]', 14))
                        feat_item = QTreeWidgetItem(device_tree, [feat_name, ''])
                        tool_tip = (feat.fget.__doc__.rstrip() if feat.fget.__doc__ else '') + (
                                    '\n\n' + feat.fset.__doc__.rstrip() if feat.fset.__doc__ else '')
                        if tool_tip != '':
                            feat_item.setData(0, Qt.ToolTipRole, tool_tip)
                        self.tree.setItemWidget(feat_item, 1, feat_widget)

                    # handle dictfeats
                    for dictfeat_name, dictfeat in device._lantz_dictfeats.items():
                        # Generate a Qt gui element for a lantz dictfeat and add docstring information as a tooltip
                        dictfeat_tree = QTreeWidgetItem(device_tree, [dictfeat_name, ''])
                        tool_tip = (dictfeat.fget.__doc__.rstrip() if dictfeat.fget.__doc__ else '') + (
                                    '\n\n' + dictfeat.fset.__doc__.rstrip() if dictfeat.fset.__doc__ else '')
                        if tool_tip != '':
                            dictfeat_tree.setData(0, Qt.ToolTipRole, tool_tip)

                        # dummy 'get' of dict feat value in order to force lantz to populate
                        # its 'subproperties' this is pretty hacky
                        for feat_key in dictfeat.keys:
                            feat = dictfeat.subproperty(getattr(device, dictfeat_name).instance, feat_key)
                            feat_widget, feat_getattr_func = self._generate_feat_widget(feat, dictfeat_name, device, dictfeat_key=feat_key)
                            self.device_to_getattr_func_mapping.append((feat_getattr_func, dictfeat_name, device, feat_key))

                            feat_widget.setFont(QFont('Helvetica [Cronyx]', 14))
                            feat_item = QTreeWidgetItem(dictfeat_tree, ['{} {}'.format(dictfeat_name, feat_key), ''])
                            self.tree.setItemWidget(feat_item, 1, feat_widget)
                            if feat_key == dictfeat.keys[-1]:
                                # connect the feat_getattr() to be called when the feat changes using
                                # pimpmyclass "ObservableProperty" mixin
                                def dictfeat_getattr_func(df_tree, df_keys, getattr_func, value, old_value, key):
                                    w = self.tree.itemWidget(df_tree.child(df_keys.index(key)), 1)
                                    getattr_func(value, old_value, widget=w)

                                # we have to use a partial here because PySignal and RPyC don't
                                # play nicely if you .connect() a lambda or other function / method
                                # to PySignal
                                feat_getattr_func.__name__ = 'InstrumentManager_getattr_func'
                                getattr_partial = functools.partial(dictfeat_getattr_func, dictfeat_tree, dictfeat.keys, feat_getattr_func)
                                getattr_partial = functools.update_wrapper(getattr_partial, feat_getattr_func)
                                getattr(device, dictfeat_name + '_changed').connect(getattr_partial)

                    # handle actions
                    action_tree = None
                    for action_name, action in device._lantz_actions.items():
                        # actions that shouldn't be added to the GUI:
                        # default lantz actions, duplicated _async actions, and
                        # any actions with parameters other than class object
                        ignore_actions = ['initialize', 'finalize', 'update', 'refresh']
                        if action_name in ignore_actions or '_async' in action_name or len(inspect.getfullargspec(action._func).args) > 1:
                            continue
                        if not action_tree:
                            action_tree = QTreeWidgetItem(device_tree, ['Actions', ''])

                        action_widget = self._generate_action_widget(device, action, action_name)

                        # set formatting and add docstring information as a tooltip
                        action_widget.setFont(QFont('Helvetica [Cronyx]', 14))
                        action_item = QTreeWidgetItem(action_tree, [action_name, ''])
                        tool_tip = action.__doc__.rstrip() if action.__doc__ else None
                        if tool_tip:
                            action_item.setData(0, Qt.ToolTipRole, tool_tip)
                        self.tree.setItemWidget(action_item, 1, action_widget)
                except Exception as exc:
                    logger.error(exc)
                    # some error has occurred loading the device attributes
                    device_tree.setText(0, 'ERROR: {}'.format(device_name))
                    device_tree.setBackground(0, QColor(255, 0, 0, 150))
                    pass


    def _generate_feat_widget(self, feat, feat_name, device, dictfeat_key=None):
        """Generate a Qt GUI element for a lantz feat/dictfeat"""

        if dictfeat_key:
            ## val = dictfeat.subproperty(getattr(device, feat_name).instance, dictfeat_key)
            # val = getattr(device, feat_name)[dictfeat_key]  # .__getitem__(dictfeat_key)
            feat_value = getattr(device, feat_name)[dictfeat_key]
            # if lantz has a function pointer in df.fset, then it is writeable
            read_only = False if getattr(device, feat_name).df.fset else True
        else:
            feat_value = getattr(device, feat_name)
            read_only = False if feat.fset else True
        
        # the lantz feat has only a specific set of allowed values
        # so we make a dropdown box
        if feat._config['values']:
            if read_only:
                # use a lineedit instead of a combo box if it's read only
                widget = QLineEdit()
                widget.setText(str(feat_value))
                widget.setReadOnly(True)
                def getattr_func(value, old_value, widget=widget):
                    widget.setText(str(value))
                    widget_color = widget.palette().base().color().name()  # 'base: #191919'
                    widget.setStyleSheet('background-color: {}'.format(UPDATE_HEX_COLOR))
                    QTimer.singleShot(UPDATE_DISPLAY_TIME, lambda: widget.setStyleSheet('selection-background-color: {}'.format(widget_color)))
            else:
                widget = QComboBox()
                disable_widget_scroll_wheel_event(widget)
                # dictionary mapping the possible lantz values to str(values)
                # e.g. {'True' : True, 'False' : False}
                keymapping_dict = {}
                for k in iter(feat._config['values']):
                    keymapping_dict[str(k)] = k

                # add the possible values to the dropdown list
                widget.addItems(keymapping_dict.keys())
                widget.setCurrentIndex(list(keymapping_dict.values()).index(feat_value))

                # callback function for when the user changes the dropdown selection
                def setattr_func(value, key=None):
                    if key:
                        getattr(device, feat_name)[key] = keymapping_dict[widget.currentText()]
                    else:
                        setattr(device, feat_name, keymapping_dict[widget.currentText()])
                # call setattr_func() when the combo box value is set from the GUI
                widget.activated.connect(functools.partial(setattr_func, key=dictfeat_key))
                def sizeHint(self):
                    return ROW_QSIZE
                widget.sizeHint = sizeHint.__get__(widget, QComboBox)

                # callback function to modify the GUI when the the feat is changed externally
                def getattr_func(value, old_value, widget=widget):
                    # because the _changed is shared for all keys of the same dictfeat,
                    # we have to check to see if this was the key that was actually changed
                    widget.setCurrentIndex(list(keymapping_dict.values()).index(value))
                    widget_color = widget.palette().alternateBase().color().name()  # 'alternatebase: #353535'
                    widget.setStyleSheet('background-color: {}'.format(UPDATE_HEX_COLOR))
                    QTimer.singleShot(UPDATE_DISPLAY_TIME, lambda: widget.setStyleSheet('selection-background-color: {}'.format(widget_color)))

        # the lantz feat is some sort of numerical value, so we will
        # generate a SpinBox (number entry with increment / decrement arrow keys)
        elif isinstance(feat_value, (int, float, Q_)):
            # arguments for the SpinBox
            optional_args = {}
            
            if feat._config['units']:
                if isinstance(feat_value, Q_):
                    base_units = infer_base_unit(feat_value)
                    base_units_str = '{0.units:~}'.format(Q_(1, base_units))
                    optional_args['suffix'] = base_units_str
                    optional_args['siPrefix'] = True
                else:
                    raise Exception('Didn\'t think this could happen... '
                        'the value of a lantz feat with units isn\'t a Q_')

            if feat._config['limits']:
                if len(feat._config['limits']) == 1:
                    # only min or only max was specified e.g.
                    # (,max) or (min,)
                    try:
                        optional_args['min'] = feat._config['limits'][0]
                    except IndexError:
                        optional_args['max'] = feat._config['limits'][1]
                else:
                    # (min, max) was specified
                    optional_args['bounds'] = feat._config['limits']

            # stepping strategy (specific settings needed to get step_widget working correctly)
            optional_args['dec'] = False
            optional_args['step'] = 1
            optional_args['minStep'] = 1e-6
            # number of decimal places to display
            optional_args['decimals'] = 6
            optional_args['compactHeight'] = False
            if isinstance(feat_value, int):
                optional_args['int'] = True
                optional_args['minStep'] = 1
                # optional_args['decimals'] = 10

            if read_only:
                pyqt_widget = pyqt_LineEdit(optional_args.get('suffix', ''), optional_args.get('siPrefix', False))
            else:
                pyqt_widget = pyqtgraph_SpinBox(**optional_args)
                disable_widget_scroll_wheel_event(pyqt_widget)
                def sizeHint(self):
                    return QSize(89, 23)
                pyqt_widget.sizeHint = sizeHint.__get__(pyqt_widget, pyqtgraph_SpinBox)
                pyqt_widget.setFont(QFont('Helvetica [Cronyx]', 14))

            if isinstance(feat_value, Q_):
                pyqt_widget.setValue(feat_value.to(base_units).m)
            else:
                pyqt_widget.setValue(feat_value)
            
            if feat._config['units']:
                # callback function for when the user changes the value from the GUI
                def setattr_func(value):
                    setattr(device, feat_name, Q_(pyqt_widget.value(), base_units))
                # callback function to modify the GUI when the the feat is changed externally
                def getattr_func(value, old_value, widget=pyqt_widget):
                    if isinstance(value, Q_):
                        pyqt_widget.setValue(value.to(base_units).m)
                    else:
                        pyqt_widget.setValue(value)
                    widget_color = pyqt_widget.palette().base().color().name()  # 'base: #191919'
                    pyqt_widget.setStyleSheet('background-color: {}'.format(UPDATE_HEX_COLOR))
                    QTimer.singleShot(UPDATE_DISPLAY_TIME, lambda: pyqt_widget.setStyleSheet('selection-background-color: {}'.format(widget_color)))
            else:
                def setattr_func(value):
                    setattr(device, feat_name, pyqt_widget.value())
                def getattr_func(value, old_value, widget=pyqt_widget):
                    pyqt_widget.setValue(value)
                    widget_color = pyqt_widget.palette().base().color().name()  # 'base: #191919'
                    pyqt_widget.setStyleSheet('background-color: {}'.format(UPDATE_HEX_COLOR))
                    QTimer.singleShot(UPDATE_DISPLAY_TIME, lambda: pyqt_widget.setStyleSheet('selection-background-color: {}'.format(widget_color)))

            if read_only:
                widget = pyqt_widget
            else:
                pyqt_widget.sigValueChanged.connect(setattr_func)

                # text that says 'step'
                step_label = QLabel()
                step_label.setText('step:')

                # editable text box where the user can enter how much the feat should step by when
                # pressing the increment/decrement arrows
                step_widget = QLineEdit()
                step_widget.setFixedSize(73, 23)
                step_widget.setFont(QFont('Helvetica [Cronyx]', 14))
                if isinstance(feat_value, Q_):
                    # set the default step units to be whatever the units are for the lantz feat
                    units_str = '{0.units:~}'.format(feat_value)
                    step_widget.setText('1 ' + units_str)
                    def set_step_func(value):
                        try:
                            new_step = Q_(value).to(base_units).m
                        except Exception as exc:
                            raise InstrumentManagerError(f'The value entered as the step [{value}] for feat [{feat_name}] couldn\'t be interpretted as a valid value with units [{base_units}]', exception=exc) from None
                        pyqt_widget.setOpts(step=new_step)
                elif isinstance(feat_value, int):
                    step_widget.setText('1')
                    def set_step_func(value):
                        try:
                            new_step = int(base_units)
                        except Exception as exc:
                            raise InstrumentManagerError(f'The value entered as the step [{value}] for feat [{feat_name}] couldn\'t be interpretted as a valid int', exception=exc) from None
                        pyqt_widget.setOpts(step=new_step)
                elif isinstance(feat_value, float):
                    step_widget.setText('1.0')
                    def set_step_func(value):
                        try:
                            new_step = float(base_units)
                        except Exception as exc:
                            raise InstrumentManagerError(f'The value entered as the step [{value}] for feat [{feat_name}] couldn\'t be interpretted as a valid float', exception=exc) from None
                        pyqt_widget.setOpts(step=new_step)
                else:
                    raise InstrumentManagerError('')
                # update the spinbox step size
                set_step_func(step_widget.text())
                step_widget.textChanged.connect(set_step_func)

                # wrap widgets in a horizontal layout and widget to format column because
                # a step_widget is being used to specify step size
                layout = QHBoxLayout()
                layout.setContentsMargins(0, 0, 0, 0)
                layout.addWidget(pyqt_widget)
                layout.addSpacing(29)
                layout.addWidget(step_label)
                layout.addWidget(step_widget)
                widget = QWidget()
                widget.setLayout(layout)

        # the lantz feat is a string, so we will just make a text box
        elif isinstance(feat_value, str):
            widget = QLineEdit()
            widget.setText(feat_value)
            def setattr_func(value):
                setattr(device, feat_name, widget.text())

            if read_only:
                widget.setReadOnly(True)
            else:
                widget.textChanged.connect(setattr_func)

            def getattr_func(value, old_value, widget=widget):
                widget.setText(value)
                widget_color = widget.palette().base().color().name()  # 'base: #191919'
                widget.setStyleSheet('background-color: {}'.format(UPDATE_HEX_COLOR))
                QTimer.singleShot(UPDATE_DISPLAY_TIME, lambda: widget.setStyleSheet('selection-background-color: {}'.format(widget_color)))

        # some unknown type - make a readonly text box containing str(feat)
        else:
            widget = QLineEdit()
            widget.setText(str(feat))
            widget.setReadOnly(True)
            def getattr_func(value, old_value, widget=widget):
                widget.setText(str(value))
                widget_color = widget.palette().base().color().name()  # 'base: #191919'
                widget.setStyleSheet('background-color: {}'.format(UPDATE_HEX_COLOR))
                QTimer.singleShot(UPDATE_DISPLAY_TIME, lambda: widget.setStyleSheet('selection-background-color: {}'.format(widget_color)))

        return widget, getattr_func

    def _generate_action_widget(self, device, action, action_name):
        """Generate a Qt gui element for a lantz action."""
        action_button = QPushButton(action_name, self.tree)
        action_button.setFont(QFont('Helvetica [Cronyx]', 12))
        def sizeHint(self):
            return ROW_QSIZE
        action_button.sizeHint = sizeHint.__get__(action_button, QPushButton)

        def action_func():
            getattr(device, action_name)()
        action_button.clicked.connect(action_func)

        return action_button


if __name__ == '__main__':
    import logging
    import sys
    from PyQt5.QtCore import Qt
    from nspyre.gui import NSpyreApp
    from nspyre.misc import nspyre_init_logger

    nspyre_init_logger(logging.INFO)

    logger.info('starting Instrument Manager...')
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp([sys.argv])
    pyqtgraph_connectCleanup()
    config_path = load_meta_config(CLIENT_META_CONFIG_PATH)
    with InservGateway(config_path) as isg:
        inserv_window = InstrumentManagerWindow(isg)
        app.exec()
    sys.exit()



"""Entry point for instrument server CLI"""
import argparse
import logging

from pathlib import Path

from nspyre.errors import ConfigurationError
from nspyre.definitions import CLIENT_META_CONFIG_PATH, SERVER_META_CONFIG_PATH
from nspyre.misc import nspyre_init_logger
from .config_files import (meta_config_add, meta_config_enabled_idx,
                           meta_config_files, meta_config_remove, meta_config_set_enabled_idx)

THIS_DIR = Path(__file__).parent
DEFAULT_LOG = THIS_DIR / 'config.log'


def main():
    # parse command-line arguments
    arg_parser = argparse.ArgumentParser(prog='nspyre-config',
                                         description='Set the nspyre configuration files')
    arg_parser.add_argument('-a', '--add-config', nargs='+',
                            default=None,
                            help='permanently add a configuration '
                                 'file(s) to the list to be imported on startup')
    arg_parser.add_argument('-d', '--delete-config', nargs='+',
                            default=None,
                            help='remove a configuration file(s) from '
                                 'the list to be imported on startup - pass either '
                                 'the file path or its index')
    arg_parser.add_argument('-l', '--list-configs',
                            action='store_true',
                            help='list the configuration files to be '
                                 'imported on startup')
    arg_parser.add_argument('-q', '--quiet',
                            action='store_true',
                            help='disable logging')
    arg_parser.add_argument('-s', '--set-config',
                            default=None,
                            help='set the active config file that will be used '
                                 'by nspyre')
    arg_parser.add_argument('-v', '--verbosity',
                            default='info',
                            help='the verbosity of logging - options are: '
                                 'debug, info, warning, error')
    arg_parser.add_argument('client_or_inserv', help='pass [client] to modify the client'
                                                     'configuration files, [inserv] for the instrument server')
    cmd_args = arg_parser.parse_args()

    # configure server logging behavior
    if not cmd_args.quiet:
        if cmd_args.verbosity.lower() == 'debug':
            log_level = logging.DEBUG
        elif cmd_args.verbosity.lower() == 'info':
            log_level = logging.INFO
        elif cmd_args.verbosity.lower() == 'warning':
            log_level = logging.WARNING
        elif cmd_args.verbosity.lower() == 'error':
            log_level = logging.ERROR
        else:
            raise Exception('Logging level [{}] not recognized.'.format(cmd_args.verbosity)) from None
        nspyre_init_logger(log_level)

    if cmd_args.client_or_inserv == 'client':
        meta_config_path = CLIENT_META_CONFIG_PATH
    elif cmd_args.client_or_inserv == 'inserv':
        meta_config_path = SERVER_META_CONFIG_PATH
    else:
        raise ConfigurationError('expected either [client] or [inserv]')

    if cmd_args.add_config:
        # the user asked us to add config files to the meta-config
        meta_config_add(meta_config_path, cmd_args.add_config)
    if cmd_args.delete_config:
        # the user asked us to remove config files from the meta-config
        meta_config_remove(meta_config_path, cmd_args.delete_config)
    if cmd_args.set_config:
        # the user asked to change the enabled config
        meta_config_set_enabled_idx(meta_config_path, cmd_args.set_config)
    if cmd_args.list_configs:
        # the user asked us to list the config files from the meta-config
        files = meta_config_files(meta_config_path)
        idx = meta_config_enabled_idx(meta_config_path)
        if files:
            for i in range(len(files)):
                print('{} {}: {}'.format('*' if i == idx else ' ', i, files[i]))
        else:
            print('no configuration files exist - use --add-config to add files')
